//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//


#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <qstring.h>
#include <qimage.h>

#include "asyncLoader.h"


void*
callAsyncLoaderGo(void *arg)
{
	((asyncLoader*)arg)->go();
	return (void*)0;
}


asyncLoader::asyncLoader()
{
	pthread_cond_init(&goCond, NULL);
	pthread_mutex_init(&goCondMutex, NULL);
	pthread_mutex_init(&lockMutex, NULL);
	*imagePath=0;
	img=0;
	timeToDie=false;
	loadNow=false;
	loadingInProgress=false;
}


asyncLoader::~asyncLoader()
{
	if (img) delete img;
}


void
asyncLoader::go()
{
	QImage *newImage;

	pthread_mutex_lock(&goCondMutex);
	while(1) {
		while(!loadNow) { // test the condition predicate
			// Be safe in case of spurious wakeups
			pthread_cond_wait(&goCond, &goCondMutex);
			// Also: remember only the unlock and wait is atomic, the lock after the
			// condition is signaled can (and does!) happen after a context switch
		}
		loadNow=false; // reset the predicate
		lock();
		if (timeToDie) {
			unlock();
			return;
		}
		if (img!=0) {
			delete img;
			img=0;
			//puts("go: unharvested image");
		}
		unlock();

		newImage=imageLoad(imagePath);

		lock();
		img=newImage;
		loadingInProgress=false; // loading has finished
		unlock();
	}
}


void
asyncLoader::lock()
{
	pthread_mutex_lock(&lockMutex);
}


void
asyncLoader::unlock()
{
	pthread_mutex_unlock(&lockMutex);
}


bool
asyncLoader::loading()
{
	bool res;
	lock();
	res=loadingInProgress;
	unlock();
	return res;
}


void
asyncLoader::loadImage(const char *path)
{
	//printf("asyncLoader::loadImage %s\n",path.data());
	if (loading()) {
		puts("#### bad: loadImage called while loading");
	}
	// tell the loader the image name
	lock();
	strncpy(imagePath, path, IMAGEPATHLEN);
	loadingInProgress=true;
	unlock();

	// Signal the loader to load a new image
	pthread_mutex_lock(&goCondMutex);
	loadNow=true; // set condition predicate
	pthread_cond_signal(&goCond);
	pthread_mutex_unlock(&goCondMutex);
}


void
asyncLoader::getLoadedImage(QImage **anImage)
{
	if (loading())
		puts("getLoadedImage:: loading while still");
	lock();
	if (*imagePath==0) {
		puts("getLoadedImage: when no image loading");
		unlock();
		return;
	}
	if (img==0) {
		puts("getLoadedImage: when image still loading");
		unlock();
		return;
	}
	*imagePath=0;

	if (anImage)
		*anImage=img;
	else // discard image if anImage == 0
		if (img)
			delete img;

	img=0;
	unlock();
}


void
asyncLoader::signalFinish()
{
	lock();
	timeToDie=true;
	unlock();
	pthread_mutex_lock(&goCondMutex);
	loadNow=true;
	pthread_cond_signal(&goCond);
	pthread_mutex_unlock(&goCondMutex);
}


// If an image is loading this function halts the process  until the load
// is finished. If no image is loading it returns immediately
void asyncLoader::waitForLoad()
{
	while(loading()) {
		pthread_mutex_lock(&goCondMutex);
		if (loading()) {
			usleep(1000);
			lock();
			printf("prob: name=%s img=%p loadNow=%d loadingInProgress=%d\n",imagePath, img, loadNow, loadingInProgress);
			unlock();
		}
		pthread_mutex_unlock(&goCondMutex);
	}
}


// Returns a pointer to a QImage object. Note that the user is responsible
// for deleting the object

QImage*
imageLoad(QString file)
{
	QImage *img=new QImage;

	bool loaded=img->load(file);
	if (!loaded || img->isNull()) { // attempt to load by extension
		int i=file.findRev('.');
		if (i!=-1) {
			QString extn=file.mid(i+1,file.length()).upper();
			if (extn=="TGA") {
				img->load(file, extn);
			}
		}
	}
	return img;
}
