//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//

#include "imageCache.h"
#include "selector.h"
#include <klocale.h>
// #include <X11/Xlib.h>
// #include <X11/Xutil.h>


// Note: this class ISN'T fast, generic, versatile or suited to large caches
// It's only really suited for this purpose

// This ain't no ordinary cache
// Items are ALWAYS inserted so the user should check to see if there is space
// before adding each item. Items inserted at higher priorities may delete
// items at lower priorities to try and keep within maxCost

// Cache class for images

imageCacheItem::imageCacheItem(selectorItem* itm, QPixmap pix, int priority)
{
	item=itm;
	image=pix;
	pri=priority;
	size=image.size();
	cost=image.width() * image.height() * image.depth()/8;
	ximg=0;

	item->setText(4, i18n("Yes"));
	if ((image.width()==0) || (image.height()==0))
		item->setText(3,"-");
	else
		item->setText(3,QString().sprintf("%dx%d",image.width(), image.height()));
 	if (item->text(2) == "")
 		item->setText(2,QImage::imageFormat(item->filePath()));
}


imageCacheItem::~imageCacheItem()
{
	//printf("Cache removed: %s\n",item->fileName().data());
	item->setText(4, i18n("No"));
}


void
imageCacheItem::describe()
{
	printf("Cache key : %-10s (cost=%6d) (pri=%d)\n", item->fileName().data(),
				 cost, pri);
}


imageCache::imageCache()
{
	maxCostNum=4000000; // Default cache size
	totalCostNum=0;
	cache.setAutoDelete(true);
	visual=rgb888x;
	//visual=rgb565;
}


// newPri==0 implies don't change the items priority

imageCacheItem *
imageCache::find(selectorItem* key, int newPri, bool internal)
{
	int idx=0;
	imageCacheItem *cItem;
	QListIterator<imageCacheItem> it(cache);
	while(it.current()) {
		cItem=it.current();
		if (key==cItem->item) {
			if (newPri!=0) {
				if (newPri<cItem->pri) // Items don't get worse in priority
					cItem->pri=newPri;
				cache.take(idx);
				cache.insert(0, cItem);
			}
			return(cItem);
		}
		idx++;
		++it;
	}
	return(0);
}


bool
imageCache::insert(selectorItem* item, QPixmap pix, int pri)
{
	if (find(item, 0, true)) {
		printf("imageCache::insert - duplicate entry %s (ignored)\n", item->fileName().data());
		return(false);
	}
	imageCacheItem *newItem = new imageCacheItem(item, pix, pri);
	while((maxCostNum-totalCostNum)<newItem->cost) {
		if (!makeSpace(pri))
			break; // Can't make any more space, add it now anyway
	}
	cache.append(newItem);
	totalCostNum+=newItem->cost;

	//printf("totalCostNum=%d\n",totalCostNum);

	return(true);
}


void
imageCache::setMaxCost(int newMaxCos)
{
	if (newMaxCos<1)
		newMaxCos=1;

	while(totalCostNum>=newMaxCos)
		makeSpace(1);

	maxCostNum=newMaxCos;
}


void
imageCache::describe()
{
	QListIterator<imageCacheItem> it(cache);
	while(it.current()) {
		it.current()->describe();
		++it;
	}
	printf("totalCost=%d\nmaxCost=%d\n",totalCostNum,maxCostNum);
}


int
imageCache::freeSpace(int pri=-1)
{
	// gives the free space for an item which will load at priority pri
	// ie if you clear out items with priority>=pri. or the absolute space
	// left if pri=-1

	if (pri==-1)
		return(maxCostNum-totalCostNum);
	
	int priCost=0;
	QListIterator<imageCacheItem> it(cache);
	while(it.current()) {
		if (it.current()->pri<pri)
			priCost+=it.current()->pri;
		++it;
	}

	return(maxCostNum-priCost);
}


bool
imageCache::makeSpace(int maxPri=1)
{
	//printf("make space %d",maxPri);
	imageCacheItem *item=cache.first();
	if (item==0) {
		return(false);
	}
	int idx=0, killidx=0;
	int thePri=0;
	QListIterator<imageCacheItem> it(cache);
	
	// find the item with the lowest priority (highest priority number)
	while(it.current()) {
		if (it.current()->pri>=thePri) {
			thePri=it.current()->pri;
			item=it.current();
			killidx=idx;
		}
		idx++;
		++it;
	}

	if (thePri>=maxPri) {
		totalCostNum-=item->cost;
		cache.remove(killidx);
		return(true);
	}
	return(false);
}


bool
imageCache::removeItem(selectorItem *item)
{
	int idx=-1, i=0;
	QListIterator<imageCacheItem> it(cache);
	while(it.current()) {
		if (it.current()->item==item) {
			idx=i;
			break;
		}
		++it;
		i++;
	}

	if (idx!=-1) {
		totalCostNum-=cache.at(idx)->cost;
		cache.remove(idx);
		return true;
	}
	return false;
}


void
imageCache::clear()
{
	//puts("--- clearing cache");
	cache.clear();
	totalCostNum=0;
}


void
imageCache::flattenPriorities(int newPri)
{
	// makes all the priorities <newPri to newPri but maintains the order
	// in terms of priorities

	for(int pri=1; pri<newPri; pri++) {
		int idx=0;
		QListIterator<imageCacheItem> it(cache);
		while(it.current()) {
			imageCacheItem *curr=it.current();
			++it;
			if (curr->pri == pri) {
				cache.take(idx);
				cache.append(curr);
				curr->pri=newPri;
				idx--;
			}
			idx++;
		}
	}
}


void
imageCache::setPixmap(QPixmap pix, selectorItem *itm)
{
// 	XImage *xi;
// 	imageCacheItem *cacheItem=theImageCache->find(itm, 1, true);

// 	if (cacheItem) {
// 		if (visual==unknown) {
// 			pix.convertFromImage(cacheItem->image);
// 		} else {
// 			puts("****fastpix****");
// 			*pix=QPixmap(cacheItem->size);
// 			xi=XCreateImage(pix->x11Display(), pix->x11Visual(), pix->x11Depth(),
// 											ZPixmap, 0, (char *)cacheItem->ximg,
// 											cacheItem->size.width(), cacheItem->size.height(),
// 											8, 0);
// 			//printf("ximage: bytes_per_line=%d\n",xi->bytes_per_line);

// 			QTime s;
// 			s.start();
// 			XPutImage(pix->x11Display(), pix->handle(), qt_xget_readonly_gc(),
// 								xi, 0,0, 0,0, cacheItem->size.width(),
// 								cacheItem->size.height());
// 			printf("*** elapsed ms=%d\n",s.elapsed());
// 			xi->data=0;
// 			XDestroyImage(xi);
// 		}
// 	} else
// 		*pix=QPixmap();
}

void
imageCache::convertImageToPixmapData(imageCacheItem *item)
{
// 	if (item->image.depth()==8)
// 		item->img=item->img.convertDepth(32);

// 	switch(visual) {
// 	case rgb565:
// 		{
// 			//puts("*** RGB565");
// 			item->ximg=(uchar *)malloc(item->img->width()*item->img->height()*2);
// 			ushort s;
// 			ushort *ptr=(ushort *)item->ximg;
// 			uchar *qimg=item->img->bits();
// 			for(int y=0;y<item->img->height();y++)
// 				for(int x=0;x<item->img->width();x++) {
// 					s=0;
// 					s|=(*qimg++ & 248)<<8;
// 					s|=(*qimg++ & 252)<<3;
// 					s|=(*qimg++)>>3;
// 					qimg++;
// 					*ptr++=s;
// 				}
// 			*(item->img)=QImage();
// 		}
// 		break;
		
// 	case rgb888x:
// 		{
// 			//puts("*** RGB888");
// 			item->ximg=(uchar *)malloc(item->img->width()*item->img->height()*4);
// 			uchar *ptr=(uchar *)item->ximg;
// 			uchar *qimg=item->img->bits();
// 			for(int y=0;y<item->img->height();y++)
// 				for(int x=0;x<item->img->width();x++) {
// 					*ptr++=*(qimg+2);
// 					*ptr++=*(qimg+1);
// 					*ptr++=*(qimg+0);
// 					*ptr++;
// 					qimg+=4;
// 				}
// 			*(item->img)=QImage();
// 		}
// 		break;
			
// 	default:
// 		puts("convertImageToPixmapData called with unknown visual");
// 		break;
// 	}
}
