//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//

#include <dirent.h>
#include <stdlib.h>

#include "kdirview.h"
#include <qdir.h>
#include <qpixmap.h>
#include <qheader.h>
#include <qaccel.h>
#include <qdragobject.h>
#include <kiconloader.h>
#include <klocale.h>
#include "kapp-safe.h"
#include "misc.h"


QPixmap *KDirViewItem::closedIcon=0, *KDirViewItem::openIcon=0,
	*KDirViewItem::indexIcon=0, *KDirViewItem::globalIndexIcon=0,
	*KDirViewItem::bothIndexIcon=0, *KDirViewItem::linkIcon=0;


KDirViewItem::KDirViewItem( KDirViewItem * parent, const char *name )
    : QListViewItem( parent ), f(name)
{
	p = parent;
	readable = TRUE;
	init();
}


KDirViewItem::KDirViewItem( QListView * parent )
    : QListViewItem( parent )
{
	f.setName("/");
	p = 0;
	readable = TRUE;
	init();
}


KDirViewItem::KDirViewItem( QListView * parent, const char *name )
    : QListViewItem( parent )
{
	f.setName(name);
	p = 0;
	readable = TRUE;
	init();
}


void
KDirViewItem::setupIcons()
{
	closedIcon=new QPixmap(UserIcon("closed"));
	openIcon=new QPixmap(UserIcon("open"));
	linkIcon=new QPixmap(UserIcon("linkFolder"));
	indexIcon=new QPixmap(UserIcon("index"));
	globalIndexIcon=new QPixmap(UserIcon("globalIndex"));
	bothIndexIcon=new QPixmap(UserIcon("bothIndex"));
}


void
KDirViewItem::cleanUpIcons()
{
	delete closedIcon;
	delete openIcon;
	delete linkIcon;
	delete indexIcon;
	delete globalIndexIcon;
	delete bothIndexIcon;
}


void
KDirViewItem::init()
{
	QString path(fullName());
	//	puts(path.data());
	QDir theDir(path);
	if (!theDir.isReadable()) {
		readable=false;
		setExpandable(false);
		setPixmap(0,*closedIcon);
		return;
	}
	bool homeIcon=false;
	if ((path==QString(getenv("HOME"))) || theDir.exists(GLOBAL_THUMB_FILE))
		homeIcon=true;

	if (theDir.exists(THUMB_FILE))
		if (homeIcon)
			setPixmap(0,*bothIndexIcon);
		else
			setPixmap(0,*indexIcon);
	else
		if (homeIcon)
			setPixmap(0,*globalIndexIcon);
		else
			setPixmap(0,*openIcon);

	if (((KDirView*)listView())->enableSubDirCheck)
		if (hasSubDirs())
			setExpandable(true);
		else
			setExpandable(false);
	else
		setExpandable(true);
	
	// This is slower than the readdir method used by hasSubDirs(0
// 	if (theDir.entryList(QDir::Dirs)->count()==2)
// 		setExpandable(false);
// 	else
// 		setExpandable(true);
}


bool
KDirViewItem::hasSubDirs()
{
	QString path=fullName();
	DIR *dir=opendir(path);
	if (dir==0) return(false);
	
	struct dirent *ent;
	//printf("path=%s\n",path.data());
 	while((ent=readdir(dir))) {
	 	QString entry(ent->d_name);
		if (entry=="." || entry=="..") continue;
		//printf("ent=%s\n",entry.data());
		QFileInfo fi(path + "/" + entry);
		if (fi.isDir()) {
			closedir(dir);
			return(true);
		}
	}
  closedir(dir);
	return(false);
}


void
KDirViewItem::setOpen( bool o )
{
	KDirViewItem *x;

	if ( o && childCount()==0 ) {
		QString s( fullName() );
		QDir thisDir( s );
		
		const QFileInfoList * files =
			thisDir.entryInfoList(QDir::Dirs | QDir::Hidden, QDir::Unsorted);
		if ( files ) {
	    QFileInfoListIterator it( *files );
	    QFileInfo * f;
	    while( (f=it.current()) != 0 ) {
				//printf("KDirViewItem::setOpen processing %s\n",f->fileName().data());
				++it;
				if ( f->fileName() == "." || f->fileName() == ".." )
					; // nothing
				else if ( f->isSymLink() ) {
					x=new KDirViewItem( this, f->fileName());
					x->setPixmap(0,*linkIcon);
				}
				else if (f->isDir()) {
					x=new KDirViewItem( this, f->fileName() );
				}
				else { // need this?
					x=new KDirViewItem( this, f->fileName());
					x->setExpandable(false);
					printf("other: %s\n", f->fileName().data());
				}
	    }
		}
	}
	QListViewItem::setOpen( o );
}


void
KDirViewItem::setup()
{
	//    setExpandable( TRUE );
    QListViewItem::setup();
}


QString
KDirViewItem::fullName()
{
	QString s;
	if ( p ) {
		s = p->fullName();
		s.append( f.name() );
		s.append( "/" );
	} else {
		s = f.name();
	}
	return s;
}


QString
KDirViewItem::text( int column ) const
{
    if ( column == 0 )
	return f.name();
    else if ( readable )
	return "Directory";
    else
	return "Unreadable Directory";
}


void
KDirViewItem::openDir(QString path, int pos)
{
	int end=path.find('/',pos);
	if (end==-1)
		end=path.length();

	QString item=path.mid(pos,end-pos);

	if (p==0) setOpen(true);

	KDirViewItem *it=(KDirViewItem *)firstChild();

	while(it) {
		if (QString(it->text(0))==item) break;
		it=(KDirViewItem *)it->nextSibling();
	}
	if (it) {
	  if (unsigned(end+1)<path.length()) {
			it->setOpen(true);
			it->openDir(path, end+1);
		}	else {
			it->listView()->setSelected(it,true);
			it->listView()->ensureItemVisible(it);
		}
	}
}

void
KDirViewItem::paintCell(QPainter *p, const QColorGroup &cg, int column,
												int width, int align)
{
	QListViewItem::paintCell(p, cg, column, width, align);
	if (this==((KDirView*)listView())->dropHighlightItem()) {
//		puts("update drag hl");
// 		QPen x=p->pen();
// 		x.setStyle(DotLine);
// 		p->setPen(x);
		p->drawRect(QRect(0,0, width,height()));
	}
}


//===================================================================


KDirView::KDirView(QWidget *parent, const char *name)
	: QListView(parent, name)
{
	init(0);
}


KDirView::KDirView(char *startDir, QWidget *parent, const char *name)
	: QListView(parent, name)
{
	init(startDir);
}


void
KDirView::init(char *startDir)
{
	emitDirSelected=true;
	
	setAcceptDrops(true);
	dropHighlight=0;

	header()->hide();
	enableSubDirCheck=true;

	KDirViewItem::setupIcons();

	setTreeStepSize(20);
	addColumn(i18n("Directory"));
	//	setRootIsDecorated(true);

	setFocusPolicy(QWidget::StrongFocus);
	viewport()->setFocusPolicy(QWidget::StrongFocus);

	if (startDir)
		setRootDir(startDir);
	
	QAccel *accels=new QAccel(this);
	accels->connectItem(accels->insertItem(Key_F5), this, SLOT(refresh()));
	connect(this, SIGNAL(selectionChanged(QListViewItem*)),
					this, SLOT(slotSelectionChanged(QListViewItem*)));
}	


KDirView::~KDirView()
{
	KDirViewItem::cleanUpIcons();
}


void
KDirView::setRootDir(QString rootDir)
{
	clear();
	enableSubDirCheck=false;
	KDirViewItem *root=new KDirViewItem(this, rootDir);
	root->setOpen(true);
	enableSubDirCheck=true;
	dirScanPos=(KDirViewItem*)firstChild()->firstChild();
	scanTimer=startTimer(20);
}


void
KDirView::setDir(QString startDir, bool emitSignal=true)
{
	//printf("setdir=%s\n", startDir.ascii());
	if (startDir.isEmpty()) {
		if (currentItem()) {
			emitDirSelected=emitSignal;
			setSelected(currentItem(), false);
			emitDirSelected=true;
		}
	}
	else {
		if (startDir.right(1)!="/") startDir.append("/");
		//printf("setDir %s\n",startDir.data());
		KDirViewItem *curr = (KDirViewItem *)currentItem();
		if (curr) {
			if (startDir==curr->fullName())
				return;	// Already there
		}
		emitDirSelected=emitSignal;
		if (firstChild())
			((KDirViewItem *)firstChild())->openDir(startDir);
		emitDirSelected=true;
	}
}


void
KDirView::resizeEvent(QResizeEvent *e)
{
	//printf("KDirView::resizeEvent %d %d\n",width(), height());
	QListView::resizeEvent(e);
}


void
KDirView::subDirScan()
{
	if (dirScanPos==0)
		return;

	//printf("checking %s\n",dirScanPos->fullName().data());
	if ((dirScanPos->isExpandable()) && (!dirScanPos->hasSubDirs())) {
		dirScanPos->setExpandable(false);
		repaintItem(dirScanPos);
	}
	dirScanPos=(KDirViewItem*)dirScanPos->itemBelow();
 	if (dirScanPos==0)
 		killTimer(scanTimer);
}


void
KDirView::timerEvent(QTimerEvent *e)
{
	if (e->timerId()==scanTimer) subDirScan();
}


void
KDirView::refresh()
{
	QStrList reopenDirs;
	QString name, last;
	// first find which dirs need to be re-opened
	KDirViewItem *item=(KDirViewItem *)firstChild();
	while((item=(KDirViewItem *)item->itemBelow())) {
		name=((KDirViewItem *)(item->parent()))->fullName();
		//printf("chk %s\n", name.data());
		if ((reopenDirs.find(name)==-1)) {
			if (last.left(name.length())==name)
				continue;
			if (last==name.left(last.length())) {
				reopenDirs.remove(reopenDirs.last());
			}
			reopenDirs.append(name);
			last=name;
		}
	}

	setUpdatesEnabled(false);
	int x=contentsX();
	int y=contentsY();

	QString currentDir=((KDirViewItem *)currentItem())->fullName();
	setRootDir(((KDirViewItem *)firstChild())->fullName());
	
	char *dirName=reopenDirs.first();
	while(dirName) {
		//printf("reload %s\n",dirName);
		setDir(dirName);
		dirName=reopenDirs.next();
	}
	setDir(currentDir);
	setContentsPos(x,y);
	setUpdatesEnabled(true);
}


void
KDirView::dragEnterEvent(QDragEnterEvent *e)
{	
	e->acceptAction(QUriDrag::canDecode(e));
}

void
xshowRect(char *name, QRect r)
{
	printf("Rect %s = (%d %d %d %d)\n",name, r.left(),r.top(), r.width(),r.height());
}


void
KDirView::dragMoveEvent(QDragMoveEvent *e)
{	
	// Only accept move or copy
	if (!((e->action()==QDropEvent::Copy) || (e->action()==QDropEvent::Move))) {
		e->ignore();
		return;
	}
	
	KDirViewItem *oldHighlight;
	if ( QUriDrag::canDecode(e) ) {
		KDirViewItem *item = (KDirViewItem*)itemAt(e->pos());
		if (item) {
			if (dropHighlight!=item) {
				oldHighlight=dropHighlight;
				dropHighlight=item;
				repaintItem(oldHighlight);
				repaintItem(dropHighlight);
			}
			//puts(item->fullName().ascii());
			e->acceptAction();
		}
	}
}

void
KDirView::dragLeaveEvent(QDragLeaveEvent *e)
{
	if (dropHighlight) {
		KDirViewItem *tmp=dropHighlight;
		dropHighlight=0;
		repaintItem(tmp);
	}
}


void
KDirView::dropEvent(QDropEvent *e)
{
	if (dropHighlight) {
		KDirViewItem *item=dropHighlight;
		dropHighlight=0;
		repaintItem(item);

		//puts("processing drop event");
	
		QStrList uris;
		if (QUriDrag::decode(e, uris)) {
			emit dropAction(uris, item->fullName(), e->action());
		}

	}
}


void
KDirView::slotSelectionChanged(QListViewItem *item)
{
	if (emitDirSelected)
		emit dirSelected(item);
}
