//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//


#include <stdio.h>
#include <stdlib.h>

#include "kgdbm.h"
#include "misc.h"


/* Some example code
	kgdbm db("theBase.dbf", false);
	db.store("andrew",28, false);
	int val=db.fetchInt("andrew");
	if (db.found())
		printf("value=%d\n",val);
	db.store("sex","male");
	QString strVal=db.fetchString("sex");
	if (db.found())
		printf("sex=%s\n",strVal.data());
	db.pack();
	printf("has andrew = %d\n",db.hasKey("andrew"));
	db.deleteItem("andrew");
	db.sync();
*/


#define barfIfDatabaseInvalid(A) \
  if (!validDataBase) {\
	  puts("Attempt to access invalide database object\n"); \
    A ;\
  }

#define EGDBMERR -9999


kgdbm::kgdbm(QString file, bool writeAccess=false, int mode=0644)
{
	int read_write= writeAccess ? GDBM_WRCREAT : GDBM_READER;
	dbfFile = gdbm_open((char*)file.data(), 0, read_write, mode, 0);

	if (dbfFile==NULL) {
		//printf("Error open database: %s\n",file.data());
		validDataBase=false;
		return;
	}
	validDataBase=true;
	lastFound=false;
	dbReadOnly=false;
}


kgdbm::~kgdbm()
{
	if (validDataBase)
		gdbm_close(dbfFile);
}


int
kgdbm::store(QString key, INT32 value, bool replace=true)
{
	int fValue=FIX_ENDIAN(value);
	barfIfDatabaseInvalid(return(EGDBMERR));
	return(storeData(key, &fValue, 4, replace));
}


int
kgdbm::store(QString key, QString value, bool replace=true)
{
	barfIfDatabaseInvalid(return(EGDBMERR));
	return(storeData(key, (char*)value.data(), value.length()+1, replace));
}


int
kgdbm::storeData(QString key, void *data, int len, bool replace=true)
{
	datum keyD, dataD;
	int flag = replace ? GDBM_REPLACE : GDBM_INSERT;

	//printf("saving rec <%s> len=%d\n",key.data(),len);

	barfIfDatabaseInvalid(return(EGDBMERR));
	
	keyD.dptr=(char*)key.data();
	keyD.dsize=key.length()+1;

	dataD.dptr=(char *)data;
	dataD.dsize=len;

	int ret=gdbm_store(dbfFile, keyD, dataD, flag);

	if (ret) {
		switch(ret) {
		case 1:
			printf("kgdbm::store - can't replace, key <%s> already in database\n",
						 key.data());
			break;
		case -1:
			printf("kgdbm::store - Database opened readonly\n");
			break;
		default:
			puts("kgdbm::store - unknown error\n");
		}
	}
	return(ret);
}


int
kgdbm::fetchInt32(QString key)
{
	datum data;

	barfIfDatabaseInvalid(return(EGDBMERR));

	if (fetchData(key, &data)) {
		int value=FIX_ENDIAN(*((int *)data.dptr));
		free(data.dptr);
		return(value);
	}
	return(EGDBMERR);
}


QString
kgdbm::fetchString(QString key)
{
	datum data;

	barfIfDatabaseInvalid(return(""));
	
	if (fetchData(key, &data)) {
		QString value=QString(data.dptr);
		free(data.dptr);
		return(value);
	}
	return(QString("NOT FOUND"));
}


int
kgdbm::fetchRec(QString key, void **data)
{
	datum dataD;

	barfIfDatabaseInvalid(return(EGDBMERR));

	if (fetchData(key, &dataD)) {
		*data=dataD.dptr;
		return(dataD.dsize);
	}
	return(-1);
}


bool
kgdbm::fetchData(QString key, datum *dataD)
{
	datum keyD;

	barfIfDatabaseInvalid(return(false));
	
	keyD.dptr=(char*)key.data();
	keyD.dsize=key.length()+1;
	
	*dataD=gdbm_fetch(dbfFile, keyD);
	if (dataD->dptr==NULL) {
		//printf("kgdbm::fetch - key %s not found\n",key.data());
		lastFound=false;
		return(false);
	}
	lastFound=true;
	return(true);
}


int
kgdbm::deleteItem(QString key)
{
	datum keyD;
	
	barfIfDatabaseInvalid(return(EGDBMERR));

	keyD.dptr=(char*)key.data();
	keyD.dsize=key.length()+1;

	int ret = gdbm_delete(dbfFile, keyD);
	if (ret == -1) printf("key <%s> not in database\n",key.data());
	return(ret);	
}


int
kgdbm::hasKey(QString key)
{
	datum keyD;
	
	barfIfDatabaseInvalid(return(EGDBMERR));

	keyD.dptr=(char*)key.data();
	keyD.dsize=key.length()+1;

	int ret = gdbm_exists(dbfFile, keyD);
	return(ret);	
}


void
kgdbm::pack()
{
	barfIfDatabaseInvalid(return);

	int ret = gdbm_reorganize(dbfFile);
	if (ret==-1)
		puts("kgdbm::pack - database opened readonly");
}


void
kgdbm::sync()
{
	barfIfDatabaseInvalid(return);

  gdbm_sync(dbfFile);
}


QString
kgdbm::firstKey()
{
	QString key;

	currentKey = gdbm_firstkey(dbfFile);
	key=(char*)currentKey.dptr;

	return key;
}


QString
kgdbm::nextKey()
{
	QString key;

	currentKey = gdbm_nextkey(dbfFile, currentKey);
	key=(char*)currentKey.dptr;

	return key;
}
