//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <klocale.h>
#include <kaboutapplication.h>
#include <kcmdlineargs.h>
#include <kurl.h>
#include <kio/jobclasses.h>
#include <kio/netaccess.h>

#include <qstring.h>
#include <qimage.h>
#include <qscrollview.h>
#include <qaccel.h>
#include <qmessagebox.h>
#include <qvaluelist.h>
#include <qdatetime.h>
#include <qmessagebox.h>

#include <config.h>

#include "selector.h"
#include "asyncLoader.h"
#include "options.h"
#include "printImageDialog.h"
#include "kconfig.h"
#include "thumbWidget.h"
#include "thumbManager.h"
#include "rapidFileConfig.h"
#include "qlistviewdrag.h"
#include "kdirview.h"
#include "viewer.h"
#include "thumbWidget.h"
#include "imageCache.h"
#include "syncCopyJob.h"

#define SB_ITEMS    1
#define SB_SCALE    2
#define SB_NAME     3
#define SB_TYPE     4
#define SB_IMG_SIZE 5
#define SB_BYTES    6
#define SB_DATE     7


bool selectorItem::fullPath;


selectorItem::selectorItem(QListView *parent, selectorItem *after,
													 QFileInfo *fi)
	: QListViewItem(parent, after)	
{
	loaded=false;
	thumbItem=0;
	setInfo(fi);

}

selectorItem::~selectorItem()
{

}

void
selectorItem::setInfo(QFileInfo *fi)
{
	fInfo=*fi;
	updateInfo();
}

void
selectorItem::updateInfo()
{
	QString tmp;

	if (fullPath)
		setText(0, fInfo.filePath());
	else
		setText(0, fInfo.fileName());

	tmp.setNum(fInfo.size());
	setText(1, tmp);
	
	setText(2, "");
	setText(3, "");
	setText(4, i18n("No"));

	tmp=fInfo.lastModified().toString();
	tmp.remove(0,4);
	setText(5, tmp);
}



QString
selectorItem::key(int column, bool ascending) const
{
	static QString theKey;
	static QDateTime epoch(QDate(1980,1,1));

	switch(column) {
	case 0:
	case 2: theKey=text(column); break;
	case 1: theKey=QString(text(column)).rightJustify(8,'0'); break;
	case 3: theKey=""; break;
	case 4: theKey=""; break;
	case 5: theKey.sprintf("%09d",epoch.secsTo(fInfo.lastModified())); break;
	}	

	//printf("col=%d key=%s\n",column,theKey.data());

	return(theKey);
}


// ========================================================================


selector::selector()
	: KTMainWindow()
{
	//opt.useAsyncLoader=false;
	
	if (opt.useAsyncLoader) {
		loader=new asyncLoader();
		pthread_create(&loaderThread, NULL, callAsyncLoaderGo,  loader);
	}
	
	// setup member variables
	treeList=0;
	treeListFirstShow=false;
	imagesLoadedFrom=directory;
	loadImageFlag=false;
	currentFileItem=0;
	dirWatchTimer=0;
	fileWatchTimer=0;
	preloadTimer=0;
	
	setCaption(i18n("PikView")+" "+VERSION+" - "+i18n("Selector"));
	getImageFormats();

	// command line file list
	KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
	QStrList listOfFiles;
	for(int i=0;i<args->count();i++)
		listOfFiles.append(args->arg(i));

	currentDir=QDir(".").absPath();
	if (args->count()==0) {    // start in lastDir only if no command line args
		if (opt.startInLastDir)  // and the option is set
			currentDir=opt.lastDir;
	}	else if (args->count()==1)
		if (QDir(args->arg(0)).exists()) {
			currentDir=QDir(args->arg(0)).absPath();
			listOfFiles.clear();
		}
	
	accels=new QAccel(this);
	accels->connectItem(accels->insertItem(Key_Escape), this, SLOT(fileQuit()));

  setMinimumSize(200,200);

	// Restore the widgets geometry if required
	if (opt.selSaveWinPos) {
		resize(opt.selWinWidth,opt.selWinHeight);
		move(opt.selWinPosX,opt.selWinPosY);
	} else
		resize(opt.selWinWidth,opt.selWinHeight);

	setUpVisual();

	// setup the image cache and look ahead loader
 	cache=new imageCache;
 	cache->setMaxCost(opt.imageCacheSize*1024);
	lookAheadNext=0;
	lookAheadDepth=0;


	// create the actual image viewer and connect the relevant signals
	theViewer = new viewer(0);

	connect(this, SIGNAL(loadImage(QPixmap,QString)),	 theViewer, SLOT(newImage(QPixmap,QString)));
	connect(theViewer, SIGNAL(imageSelectSignal(int)), this, SLOT(goImage(int)));
	connect(theViewer, SIGNAL(setSelectorVisibile(enum logicChoice)), this, SLOT(setVisible(enum logicChoice)));
	connect(theViewer, SIGNAL(viewerClosed()), this, SLOT(fileQuit()));
	connect(theViewer, SIGNAL(refileImage(QString, QString)),	this, SLOT(refileImage(QString, QString)));

	// create the selector pane as a splitter with three widgets
	// directory view, file list view and thumbnail view
	// the fileview and the thumbnail view are never shown together
	split=new QSplitter(this);
	split->setOrientation(QSplitter::Horizontal);

	// pane 1: the directory view
	treeList=new KDirView(split);
	treeList->setFocusPolicy(QWidget::StrongFocus);
	// pane 2: file list view
	fileList=new QListViewDrag(split);
	fileList->setFocusPolicy(QWidget::StrongFocus);
	fileList->viewport()->setFocusPolicy(QWidget::StrongFocus);
	fileList->setSelectionMode(QListView::Extended);
	fileList->setShowSortIndicator(true);
	// pane 3: thumbnail view
	thumbNails=new thumbWidget(split);
	thumbNails->setFocusPolicy(QWidget::StrongFocus);

	split->setResizeMode(treeList, QSplitter::KeepSize);

	//set the width of the tree view
	QValueList<int> vals;
	vals.append(opt.treeDirWidth);
	split->setSizes(vals);
	
	if (opt.fileViewType==0)
		thumbNails->hide();     // filelist view
	else
		fileList->hide();       // thumbnail view

	connect(treeList, SIGNAL(dirSelected(QListViewItem *)),this, SLOT(newDirSelected(QListViewItem *)));
	connect(treeList, SIGNAL(dropAction(QStrList&, QString, QDropEvent::Action)),
					this, SLOT(refileItems(QStrList&, QString, QDropEvent::Action)));
	
	connect(thumbNails, SIGNAL(doubleClicked(QListViewItem*)),  this, SLOT(showImageWindow(QListViewItem*)));
	connect(thumbNails, SIGNAL(itemRenamed(QIconViewItem*)),    this, SLOT(renameItem(QIconViewItem*)));
	connect(thumbNails,	SIGNAL(dropped(QDropEvent *, const QValueList<QIconDragItem> &)),
					this, SLOT(slotDropEvent(QDropEvent*)));
	
	setView(split);
	

	fileList->addColumn(i18n("Name"));
	fileList->addColumn(i18n("Size"));
	fileList->setColumnAlignment(1, AlignRight);
	fileList->addColumn(i18n("Type"));
	fileList->setColumnAlignment(2, AlignCenter);
	fileList->addColumn(i18n("Image size"));
	fileList->setColumnAlignment(3, AlignCenter);
	fileList->addColumn(i18n("Loaded"));
	fileList->setColumnAlignment(4, AlignCenter);
	fileList->addColumn(i18n("Modified"));
	//	fileList->setColumnAlignment(5, AlignRight);
	for(int i=0;i<6;i++)
			fileList->setColumnWidthMode(i, QListView::Maximum);

	fileList->setAllColumnsShowFocus(true);

	connect(fileList, SIGNAL(doubleClicked(QListViewItem *)),    this, SLOT(showImageWindow(QListViewItem*))),
	connect(fileList, SIGNAL(returnPressed(QListViewItem *)),    this, SLOT(showImageWindow(QListViewItem*)));
	connect(fileList, SIGNAL(dropped(QDropEvent *)),             this, SLOT(slotDropEvent(QDropEvent*)));
	
	QAccel *a=new QAccel(fileList);
	a->connectItem(a->insertItem(Key_Delete), this, SLOT(slotDeleteKey()));
	a=new QAccel(thumbNails);
	a->connectItem(a->insertItem(Key_Delete), this, SLOT(slotDeleteKey()));
		
	connect(theViewer, SIGNAL(newScale(float)),	this, SLOT(setStatusScale(float)));

  setupMenuBar();
  setupStatusBar();
	setStatusBarVisible(opt.showStatusBar ? on : off);
	
	if (opt.hideDirTree)
		hideDirTree();
	else
		showDirTree();

	if (opt.showSelectorAndViewer || opt.startInSelector)
		show();

	if (!listOfFiles.isEmpty()) {
		loadSelectedImages(&listOfFiles);
		if (opt.startImage==0)
			opt.startImage=1;
	} else
		loadAllImages(currentDir);
	
	// XXX this  will show a window which will get resized which is a pain
	if (opt.showSelectorAndViewer || opt.fullScreen || !opt.startInSelector)
		theViewer->show();

	if (opt.startImage)
		goImage(opt.startImage);
	else {
		theViewer->goSplashImage();
		currentFileItem=0;
	}
	
}


selector::~selector()
{
	if (opt.useAsyncLoader) {
		loader->signalFinish();
		pthread_join(loaderThread, NULL);
		delete loader;
	}
	
	delete theViewer;
	opt.saveImageCacheOptions(cache, true);
	delete cache;
	opt.saveSelectorOptions(this, true);
}

void
selector::hide()
{
	if (!isHidden()) {
		opt.selWinPosX=pos().x();
		opt.selWinPosY=pos().y();
	}
	KTMainWindow::hide();
}

void
selector::goImage(int go)
{
	if (fileList->firstChild()==0) return;

	QListViewItem *current=currentFileItem;

	if (current==0) { // fudge for no current selection
		current=fileList->firstChild();
		if ((go>0) && (go!=top)) go--;
	}
	
	switch(go) {
	case top:
		current=fileList->firstChild();
		break;
	case bottom:
		while(current->itemBelow())
			current=current->itemBelow();
		break;
	case randomPic:
		if (noItems<=1) return;
		go=rand() % noItems;
		current=fileList->firstChild();
		// NB intentional fall through
	default:
		if (go<0) {
			for(; go && current; go++)
				current=current->itemAbove();
			if (current==0) {
				opt.wrapAround ? goImage(bottom) : goImage(top);
				return;
			}
		} else {
			// can't use itemBelow from the constructor
			for(; go && current; go--)
				current=current->nextSibling();
			//current=current->itemBelow();
			if (current==0) {
				opt.wrapAround ? goImage(top) : goImage(bottom);
				return;
			}
		}
	}
	if (opt.fileViewType==0) {                // filelist view
		fileList->clearSelection();
		fileList->setSelected(current, true);
	} else {                                  // thumbnail view
		((selectorItem*)current)->thumbItem->setSelected(true);
	}
	newImageQLVI(current);
}


void
selector::loadSelectedImages(QStrList *files)
{
	QValueList<QFileInfo> fileInfoList;
	QValueList<QDir> dirList;

	QString currDir;
	QFileInfo fi;

	for(char *aFile=files->first(); aFile; aFile=files->next()) {
		fi.setFile(aFile);
		if (fi.isFile()) {
			fileInfoList.append(fi);
			if (!dirList.contains(fi.dir())) {
				dirList.append(fi.dir());
			}
		}
		else
			printf("File: %s does not exist - ignoring\n",aFile);
	}
	loadImages(&fileInfoList, &dirList, list);
}


void
selector::loadImages(QValueList<QFileInfo> *imageList, QValueList<QDir> *dirList, loadedFrom where)
{
	// Stop periodic tasks
	killTimers(); // stop the 3 timers
	discardLoadingImage();

	// Clear the current file/thumb lists
	cache->clear();
	fileList->clear();
	thumbNails->clear();
	currentFileItem=0;

	noItems=imageList->count();
	updateItemCount();

	// Set the sorting
	imagesLoadedFrom=where;
	if (imagesLoadedFrom==directory)
		fileList->setSorting(0);
	else
		fileList->setSorting(-1);
	
	// Make a list of the directories in use and open them on the dir tree
	currentDirs.clear();
	QValueList<QDir>::Iterator it2;
	for(it2=dirList->begin(); it2!=dirList->end(); ++it2) {
		currentDirs.append(QFileInfo((*it2).path()));
		treeList->setDir((*it2).path(), false);
		currentDir=(*it2).path();
	}
	if (currentDirs.count()!=1) {
		currentDir="";
		treeList->setDir("", false);
	}

	// Insert the picture files into the lists
	selectorItem::fullPath=(dirList->count()!=1);
	QValueList<QFileInfo>::Iterator it;
	
	selectorItem *sItem=0;
	for(it=imageList->begin(); it != imageList->end(); ++it) {
		sItem=new selectorItem(fileList, sItem, &(*it));
		thumbWidgetItem *tItem=new thumbWidgetItem(thumbNails, sItem);
	}
	thumbNails->sort();
	thumbNails->updateThumbNails();

	// restart the periodic tasks
	dirWatchTimer=startTimer(100);
	fileWatchTimer=startTimer(5000);
}
										
void
selector::loadAllImages(QString dir)
{
	QValueList<QFileInfo> fileInfoList;
	QValueList<QDir> dirList;
	
	QDir current(dir);
	dirList.append(current);
	
	if (current.isReadable()) {
		int filterSpec=QDir::Files | QDir::Readable | (opt.showHidden ? QDir::Hidden : 0);
		const QFileInfoList *list	=	current.entryInfoList(filterSpec, QDir::Unsorted);
		QFileInfoListIterator it( *list );
		QFileInfo *fi;
		
		while((fi=it.current())) {
			if (opt.showNonImageFiles || isImageFileName(fi->fileName()))
				fileInfoList.append(*fi);
			++it;
		}
	}
	loadImages(&fileInfoList, &dirList, directory);
}


void
selector::updateItemCount()
{
	QString items;
	items.sprintf(i18n("%d Items"),noItems);
	statusBar->changeItem(items, SB_ITEMS);
}


void
selector::getImageFormats()
{
	QStrList inpFmt=QImage::inputFormats();
	for(char *fmt=inpFmt.first(); fmt ; fmt=inpFmt.next()) {
		inputFormats.append(QString(".")+QString(fmt).upper());
		if (QString(fmt)=="JPEG")
			inputFormats.append(".JPG");
		//printf("Format: %s\n",fmt);
	}
}


bool
selector::isImageFileName(QString file)
{
	file=file.upper();
	for(char *fmt=inputFormats.first(); fmt ; fmt=inputFormats.next()) {
		if (file.right(strlen(fmt))==QString(fmt))
			return(true);
	}
	return(false);
}


// This slot is only to avoid the type mismatch warning when the filelist
// QListView (which contains only selectorItems emits its selectionChanged
// signal as a QListViewItem*

void
selector::newImageQLVI(QListViewItem *item)
{
	newImage((selectorItem*)item);
}

void
selector::newImage(selectorItem *item)
{
	QString filePath=item->filePath();
	//cache->describe();
	//printf("newImage: %s ",item->fileName().data());
	
	killTimer(preloadTimer); 		// stop image harvesting

	if (opt.useAsyncLoader && (requestedImageName==filePath)) {
		//printf("(wait for load)\n");
		loader->waitForLoad(); //wait until loaded
		harvestImage(); // put the image into the cache
	}
	
	// set priority number to zero so it can't get removed by the timerEvent
	// before the image is loaded
	imageCacheItem *cItem=cache->find(item, 1);
	if (cItem) {
		//printf("(cached)\n");
		emit loadImage(cItem->image, filePath);
	}
	else {
		// make this do a fast load
		//printf("(directloading)\n");
		QImage *img=imageLoad(filePath);
		QPixmap pix;
		pix.convertFromImage(*img);
		delete img;
		if (pix.mask()) removeMask(&pix);
		cache->insert(item, pix, 1);
		emit loadImage(pix, filePath);
	}
	setStatus(item);
	currentFileItem=item;
	
	if (opt.useAsyncLoader && opt.maxLookAhead)
		setupLookHead(item);

	//cache->describe();
	//puts("");
}


void
selector::setupLookHead(selectorItem *item)
{
	if ((opt.slideShowDir==3) && (theViewer->slideShowMode()))
		return;

	// start the lookahead process
	killTimer(preloadTimer);
	lookAheadNext=(selectorItem*)item->nextSibling();
	lookAheadDepth=0;

	// Re-prioritise items in the cache so that the lookahead process
	// avoids deleting items when will now be pre-loaded
	cache->flattenPriorities(opt.maxLookAhead+2);
	for(int look=1; (look<=opt.maxLookAhead) && item; look++) {
		cache->find(item, look);
		item=(selectorItem*)item->nextSibling();
	}
	preloadTimer=startTimer(10);
}


void
selector::newDirSelected(QListViewItem *item)
{
	QString theDir=((KDirViewItem *)item)->fullName();
	//printf("new dir=%s\n",theDir.data());
	loadAllImages(theDir);
}


void
selector::hideDirTree()
{
	opt.treeDirWidth=treeDirWidth();

	treeList->hide();
	opt.hideDirTree=true;
	view->setItemChecked(hide_dir_id, opt.hideDirTree);
	fileList->setFocus(); // XXX file list or thumbnails?
}


void
selector::showDirTree()
{
	treeList->show();
	opt.hideDirTree=false;
	view->setItemChecked(hide_dir_id, opt.hideDirTree);
	fileList->setFocus();  // XXX file list or thumbnails?

	if (!treeListFirstShow) { // wait until the treeList is first used before
		treeListFirstShow=true; // scanning all the directories
		treeList->setRootDir("/");
		if (!currentDir.isEmpty())
			treeList->setDir(currentDir);
	}
}


void
selector::toggleDirTree()
{
	if (!opt.hideDirTree)
		hideDirTree();
	else
		showDirTree();
}


void
selector::setupMenuBar()
{
  file = new QPopupMenu ();
  view = new QPopupMenu ();
  help = new QPopupMenu ();
	thumb= new QPopupMenu ();

	//  file->insertItem("&Open..."); //maybe have this for open URL
	//  file->insertSeparator();
  file->insertItem(i18n("&Settings..."), this, SLOT(options()),ALT+Key_S);
  file->insertItem(i18n("&Configure shortcuts..."), this, SLOT(configShortcuts()));
  file->insertItem(i18n("&Print..."),    this, SLOT(printImage()), CTRL+Key_P);
  file->insertSeparator();
  file->insertItem(i18n("&Quit"),        this, SLOT(fileQuit()), CTRL+Key_Q);

  view->setCheckable(TRUE);
  hide_dir_id=view->insertItem(i18n("Hide directory tree"), this, SLOT(toggleDirTree()));

  view->insertSeparator();
  showStatus_id = view->insertItem(i18n("Show status bar"), this, SLOT(toggleStatusBar()));
  view->insertSeparator();
	file_view_id=view->insertItem(i18n("File list"), this, SLOT(setFileListView()), Key_F7);
	thumb_view_id=view->insertItem(i18n("Thumbnails"), this, SLOT(setThumbNailView()), Key_F8);
	if (opt.fileViewType==0)
		view->setItemChecked(file_view_id,true);
	else
		view->setItemChecked(thumb_view_id,true);
  view->insertSeparator();
	view->insertItem(i18n("Fullscreen"), this, SLOT(goFullScreen()));

  help->insertItem(i18n("&About..."), this, SLOT(about()));
  help->insertItem(i18n("&Help..."),  this, SLOT(helpContents()), Key_F1);
	help->insertItem(i18n("&Keys..."),  this, SLOT(helpKeys()));
	
	thumb->insertItem(i18n("Create thumbnail file here..."), this, SLOT(thumbFileCreate()));
	thumb->insertItem(i18n("Delete thumbnail file here..."), this, SLOT(thumbFileDelete()));
	thumb->insertItem(i18n("Clear thumbnail file here..."),  this, SLOT(thumbFileClear()));
	thumb->insertItem(i18n("Pack thumbnail file here..."),   this, SLOT(thumbFilePack()));
	thumb->insertItem(i18n("Rebuild thumbnail file here..."),this, SLOT(thumbFileRebuild()));

  menuBar = new KMenuBar(this);
  menuBar->insertItem(i18n("&File"), file);
  menuBar->insertItem(i18n("&View"), view);
  menuBar->insertItem(i18n("&Thumbnails"), thumb);
  menuBar->insertItem(i18n("&Help"), help);
  setMenu(menuBar);

  connect(theViewer, SIGNAL(viewerDropEvent(QDropEvent *)), this, SLOT(slotDropEvent(QDropEvent *)));
	connect(theViewer, SIGNAL(deleteCurrentImage()), this, SLOT(slotDeleteCurrentImage()));
}


void
selector::setupStatusBar()
{
	statusBar = new KStatusBar(this);
	statusBar->insertItem("", SB_ITEMS, 1);
	statusBar->insertItem("", SB_SCALE, 1);
	statusBar->insertItem("", SB_NAME, 2);
	statusBar->insertItem("", SB_TYPE, 1);
	statusBar->insertItem("", SB_IMG_SIZE, 1);
	statusBar->insertItem("", SB_BYTES, 1);
	statusBar->insertItem("", SB_DATE, 2);
	setStatusBar(statusBar);
}


void
selector::fileQuit()
{
  delete this;
}


void
selector::helpContents()
{
  kapp->invokeHelp();
}

void
selector::helpKeys()
{
  kapp->invokeHelp("VIEWERKEYS");
}

void
selector::about()
{
	KAboutApplication about;
	about.exec();
}

void
selector::saveProperties(KConfig *conf)
{}

void
selector::readProperties(KConfig *conf)
{}


void
selector::setStatusBarVisible(enum logicChoice set)
{
	if (set==toggle)
		opt.showStatusBar=!opt.showStatusBar;
	else
		opt.showStatusBar=set;

  view->setItemChecked(showStatus_id, opt.showStatusBar);
  if (opt.showStatusBar)
    enableStatusBar(KStatusBar::Show);
  else
    enableStatusBar(KStatusBar::Hide);
}


void
selector::toggleStatusBar()
{
	setStatusBarVisible(toggle);
}


void
selector::setStatus(selectorItem *item)
{
	statusBar->changeItem(item->text(0),SB_NAME);
	statusBar->changeItem(item->text(1),SB_BYTES);
	statusBar->changeItem(item->text(2),SB_TYPE);
	statusBar->changeItem(item->text(3),SB_IMG_SIZE);
	statusBar->changeItem(item->text(5),SB_DATE);
}


void
selector::setStatusScale(float scaleFact)
{
	QString scale;
	scale.sprintf("%3.0f%%",scaleFact*100);
	statusBar->changeItem(scale,SB_SCALE);
}


void
selector::setVisible(enum logicChoice set)
{
	if (set==toggle)
		set=isVisible() ? off : on;

	//printf("setvisible=%d\n",set);

	switch(set) {
	case on:
		show();
		if (!opt.showSelectorAndViewer)
			theViewer->hide();
		break;
	case off:
		hide();
		if (!opt.showSelectorAndViewer)
			theViewer->show();
		break;
	case toggle: ;
	}
}


void
selector::printImage()
{
	int res;

	QPrinter prn;
	do {
		if (!QPrintDialog::getPrinterSetup(&prn))
			return;	
		printImageDialog printDlg(theViewer->currentPicture(),
															theViewer->currentFile(), &prn);
		res=printDlg.exec();
	}
	while(res==2);
}


void
selector::setFileListView()
{
	if (opt.fileViewType==0)
		return;

	opt.fileViewType=0;
	view->setItemChecked(file_view_id, true);
	view->setItemChecked(thumb_view_id, false);

	syncSelections(1);
	
	thumbNails->hide();
	fileList->show();
}


void
selector::setThumbNailView()
{
	if (opt.fileViewType==1)
		return;

	opt.fileViewType=1;
	view->setItemChecked(file_view_id, false);
	view->setItemChecked(thumb_view_id, true);

	if (treeList->width()>1)
		opt.treeDirWidth=treeList->width();

	// We rebuild the thumbnail list here so that the sort order from the
	// filelist is the same.
	thumbNails->setImageList((selectorItem *)fileList->firstChild());
	syncSelections(0);
	
	fileList->hide();
	thumbNails->show();
}

void
selector::syncSelections(int dir)
{
	QIconViewItem *item;

	if (dir==0) {
		for(item=thumbNails->firstItem(); item; item=item->nextItem())
			item->setSelected(((thumbWidgetItem*)item)->item->isSelected(), true);
	} else {
		for(item=thumbNails->firstItem(); item; item=item->nextItem())
			((thumbWidgetItem*)item)->item->setSelected(item->isSelected());
	}
}


#include <X11/Xlib.h>

void
selector::setUpVisual()
{
//   QPixmap p;
//   Display *dpy    =p.x11Display();
//   int displayDepth=p.x11Depth();
//   Visual *vis     =(Visual*)p.x11Visual();
//   bool trueColour = (vis->c_class == TrueColor);

//   visual=unknown;
//   if (trueColour) { // do they have a worthy display
//     uint red_mask  =(uint)vis->red_mask;
//     uint green_mask=(uint)vis->green_mask;
//     uint blue_mask =(uint)vis->blue_mask;

//     if ((red_mask==0xf800) && (green_mask==0x7e0) && (blue_mask==0x1f))
//       visual=rgb565;
//     if ((red_mask==0xff0000) && (green_mask==0xff00) && (blue_mask==0xff))
//       visual=rgb888x;
		
//     if (visual==unknown) {
//       puts("Unoptimized visual - want to write an optimised routine?");
//       printf("red=%8x green=%8x blue=%8x\n",red_mask,green_mask,blue_mask);
//     } else {
//       puts("Using optimized visual");
//     }
//   }
}	


void
selector::showImageWindow(QListViewItem *item)
{
	//printf("showImageWindow=%d\n",opt.showSelectorAndViewer);
	if (!opt.showSelectorAndViewer)
		setVisible(off);
	theViewer->show();
	theViewer->raise();
	newImage((selectorItem*)item);
}


void
selector::options()
{
	opt.doDialog(theViewer, this, thumbNails, cache);
}


void
selector::timerEvent(QTimerEvent *e)
{
	if ((e->timerId()!=preloadTimer) && (e->timerId()!=dirWatchTimer)
			&& (e->timerId()!=fileWatchTimer))
		puts("misc timer event");
	
	if (e->timerId()==preloadTimer) {
		if (!requestedImageName.isEmpty()) {
			//printf("timerevent:: chk for harvest %s\n", requestedImageName.data());
			if (loadImageFlag && (!loader->loading())) // load finished
				harvestImage();
			return;
		}
		
		if (lookAheadNext && (cache->freeSpace(lookAheadDepth+1)>0))
			if (!cache->find(lookAheadNext, lookAheadDepth+1)) {
				//puts("timerevent: cacheItem");
				cacheItem(lookAheadNext);
			}

		if (lookAheadNext==0) {
			killTimer(preloadTimer);
			return;
		}
		
		lookAheadNext=(selectorItem*)lookAheadNext->nextSibling();
		lookAheadDepth++;
		if ((!lookAheadNext) || (lookAheadDepth>=opt.maxLookAhead))
			lookAheadNext=0;
	}
	
	// Watches directories for changes (additions and deletions)
	if (e->timerId()==dirWatchTimer) {
		QDateTime current=QDateTime::currentDateTime();
		current.setTime(current.time().addMSecs(-current.time().msec()));
		
		if (!delayedUpdate.isNull() && delayedUpdate!=current) {
			updateFileList((imagesLoadedFrom==directory));
			delayedUpdate=QDateTime();
			return;
		}
		QValueList<QFileInfo>::Iterator it;
		for(it=currentDirs.begin(); it!=currentDirs.end(); ++it) {
			QDateTime dt=(*it).lastModified();
			(*it).refresh();
			if ((*it).lastModified()!=dt) {
				// wait until this second has passed so we get all updates during it
				delayedUpdate=current;
				return;
			}
		}
	}
	if (e->timerId()==fileWatchTimer)
		updateFileList(false);
}

void
selector::updateFileList(bool checkForAdditions)
{
	//puts("updateFileList");

	bool addedItems=false, deletedItems=false, modifiedItems=false;
	QStringList pathList;

	for(selectorItem *item=(selectorItem *)fileList->firstChild(); item;) {
		QDateTime dt=item->fInfo.lastModified();
		item->fInfo.refresh();

		if (item->fInfo.exists())
			pathList.append(item->filePath());
		else {
			selectorItem *delItem=item;
			//printf("removed %s\n",delItem->filePath().ascii());
			item=(selectorItem *)item->nextSibling();
			removeItem(delItem);
			deletedItems=true;
			continue; // ie we've already incremented item
		}
		if (dt!=item->fInfo.lastModified()) {
			//printf("changed %s\n",item->fInfo.filePath().ascii());
			cache->removeItem(item);
			item->thumbNail=QPixmap();
			if (item->thumbItem) item->thumbItem->setPixmap(QPixmap());
			item->updateInfo();
			modifiedItems=true;
		}
		item=(selectorItem *)item->nextSibling();
	}

	if (checkForAdditions) {
		if (imagesLoadedFrom!=directory) { puts("loadedFrom!=directory!!!"); }
		pathList.sort();

		QDir dir=QDir(currentDirs.first().filePath());

		int filterSpec=QDir::Files | QDir::Readable | (opt.showHidden ? QDir::Hidden : 0);
		QFileInfoListIterator updFilesIt=QFileInfoListIterator(*(dir.entryInfoList(filterSpec, QDir::Name)));

		// Check for any new files wrt pathList
		QStringList addFilesList;
		QStringList::Iterator pathListIt=pathList.begin();
		int c;
		while(updFilesIt.current()) {
			//printf("checking pathListIt=%s updFilesIt=%s\n", (*pathListIt).ascii(),(**updFilesIt).filePath().ascii());
			if (pathListIt==pathList.end())
				c=1;
			else
				c=(*pathListIt).compare((**updFilesIt).filePath());

			if (c==0) { ++pathListIt; ++updFilesIt; }
			if (c>0)  { addFilesList.append((**updFilesIt).filePath()); ++updFilesIt; }
			if (c<0)  { ++pathListIt; }
		}
		//printf("t(add search)=%d\n",t.elapsed());
		
		for(QStringList::Iterator it=addFilesList.begin(); it!=addFilesList.end(); ++it) {
			if (opt.showNonImageFiles || isImageFileName(*it)) {
				//printf("added %s\n",(*it).ascii());
				addedItems=true;
				QFileInfo fi(*it);
				selectorItem    *sItem=new selectorItem(fileList, 0, &fi);
				thumbWidgetItem *tItem=new thumbWidgetItem(thumbNails, sItem);
				noItems++;
			}
		}
	}
	
	if (deletedItems || addedItems)
			updateItemCount();

	if (deletedItems)
		thumbNails->arrangeItemsInGrid(true);

	if (addedItems || modifiedItems)
		thumbNails->updateThumbNails();
}

void
selector::discardLoadingImage()
{
	killTimer(preloadTimer);
	if (opt.useAsyncLoader && !requestedImageName.isEmpty()) {
		//puts("discardLoadingImage");
		loader->waitForLoad();     // wait until loaded
		loader->getLoadedImage(0); // discard image
		requestedImageName="";
		requestedItem=0;
		loadImageFlag=false;
	}
}

void
selector::harvestImage()
{
	QImage *requestedImage;

	if (requestedImageName.isEmpty()) {
		//puts("harvestImage:: when no image loading");
		return;
	}

	//printf("harvest %s\n",requestedImageName.data());

	// add the image just loaded by the loader thread to the cache
	loader->getLoadedImage(&requestedImage);
	// requestedImage.load(requestedImageName); // Sneakily load now if no loader
	if (requestedImage) {
		QPixmap pix;
		pix.convertFromImage(*requestedImage);
		if (pix.mask()) removeMask(&pix);
		delete requestedImage;
		cache->insert(requestedItem, pix, lookAheadDepth+1);
		//printf("Image: %s loaded and cached\n",requestedImageName.data());
	} else {
		//puts("harvestImage: load failed");
	}
	requestedImageName="";
	requestedItem=0;
	loadImageFlag=false;
}


void
selector::cacheItem(selectorItem *item)
{
	//printf("caching: %s\n",item->fileName().data());
	requestedImageName=item->filePath();
	requestedItem=item;
	loadImageFlag=true;
	loader->loadImage(requestedImageName.data());
}


void
selector::thumbFileCreate()
{
	QFileInfo fi(currentDir+THUMB_FILE);
	if (fi.exists()) {
		QMessageBox::warning(this, i18n("PikView"),
												 i18n("A thumbnail file already exists in this directory"));
		return;
	}

	int res=QMessageBox::information(this, i18n("PikView"),
																	 i18n("Create thumbnail file\n\n")+
																	 currentDir+THUMB_FILE+"?", QMessageBox::Ok,
																	 QMessageBox::Cancel);

	if (res==QMessageBox::Ok) {
		QFile tn(currentDir+THUMB_FILE);
		if (!tn.open(IO_WriteOnly))
			QMessageBox::warning(this, i18n("PikView"),
													 i18n("The file\n\n%1\n\ncannot be created.").arg(currentDir+THUMB_FILE));
		treeList->refresh();
	}
}


void
selector::thumbFileDelete()
{
	QFileInfo fi(currentDir+THUMB_FILE);
	if (fi.exists()) {
		int res=QMessageBox::information(this, i18n("PikView"),
																		 i18n("Delete thumbnail file\n\n")+currentDir+
																		 THUMB_FILE+"?", QMessageBox::Ok,
																		 QMessageBox::Cancel);
		
		// close the current database in case we're deleting it
		thumbMgr->closeDB();
		if (res==QMessageBox::Ok) {
			QFile tn(currentDir+THUMB_FILE);
			if (!tn.remove())
				QMessageBox::warning(this, i18n("PikView"),
														 i18n("The file\n\n%1\n\ncannot be removed.").arg(currentDir+THUMB_FILE));
			treeList->refresh();
		}
	}
	else {
		QMessageBox::warning(this, i18n("PikView"), i18n("There is no thumbnail file in this directory"));
		return;

	}
}


void
selector::thumbFileClear()
{
	QFileInfo fi(currentDir+THUMB_FILE);
	if (fi.exists()) {
		int res=QMessageBox::information(this, i18n("PikView"),
																		 i18n("Clear thumbnail file\n\n")+
																		 currentDir+THUMB_FILE+"?",
																		 QMessageBox::Ok, QMessageBox::Cancel);
		
		// close the current database in case we're clearing it
		thumbMgr->closeDB();
		if (res==QMessageBox::Ok) {
			QFile tn(currentDir+THUMB_FILE);
			if (!tn.open(IO_WriteOnly | IO_Truncate))
				QMessageBox::warning(this, i18n("PikView"),
														 i18n("The file\n\n%1\n\ncannot be cleared.").arg(currentDir+THUMB_FILE));
			treeList->refresh();
		}
	}
	else {
		QMessageBox::warning(this, i18n("PikView"), i18n("There is no thumbnail file in this directory"));
		return;
		
	}
}


void
selector::thumbFilePack()
{
	QString database, dirBase;
	database=thumbMgr->findThumbNailFile(currentDir, dirBase);
	if (database.isEmpty()) {
		QMessageBox::warning(this, i18n("PikView"),
												 i18n("Cannot find thumbnail file for this directory"));
		return;
	}
	int res=QMessageBox::information(this, i18n("PikView"),
																	 i18n("Pack thumbnail file\n\n")+
																	 database+"?",
																	 QMessageBox::Ok, QMessageBox::Cancel);
	
	if (res!=QMessageBox::Ok)
		return;

	//printf("pack: file=%s\n",database.data());
	thumbMgr->packDatabase(database);
}


void
selector::thumbFileRebuild()
{
	QString database, dirBase;
	database=thumbMgr->findThumbNailFile(currentDir, dirBase);
	if (database.isEmpty()) {
		QMessageBox::warning(this, i18n("PikView"), i18n("Cannot find thumbnail file for this directory"));
		return;
	}
	int res=QMessageBox::information(this, i18n("PikView"),
																	 i18n("Rebuild thumbnail file\n\n")+
																	 database+"?",
																	 QMessageBox::Ok, QMessageBox::Cancel);
	
	if (res!=QMessageBox::Ok)
		return;

	//printf("rebuild: file=%s\n",database.data());
	thumbMgr->rebuildDatabase(database);
}


void
selector::resizeEvent(QResizeEvent *e)
{
	//puts("selector::resizeEvent");
// 	opt.selWinWidth=width(); //XXX remove
// 	opt.selWinHeight=height();
	KTMainWindow::resizeEvent(e);
}


void
selector::moveEvent(QMoveEvent *)
{
	// XXX remove
// 	static int blockEvents=2;

// 	if (blockEvents) {
// 		blockEvents--;
// 		return;
// 	}

// 	opt.selWinPosX=pos().x();
// 	opt.selWinPosY=pos().y();
}


void
selector::goFullScreen()
{
	theViewer->setFullScreen(on);
}


void
selector::refileImage(QString source, QString dest)
{
	if ((!QFile(source).exists()) || (!QDir(dest).exists())) {
		puts("refileImage: source or dest doesn't exist\n");
		return;
	}

	selectorItem *current=currentFileItem;

	if (current==0)
		return;
	if (current->filePath()!=source) {
		puts("refileImage: source != current");
		return;
	}

	QString mess;
	mess.sprintf("%s -> %s",current->fileName().data(), dest.data());
	
	QStrList srcUris;
	srcUris.append(source);
	refileItems(srcUris, dest, QDropEvent::Move);
	
	if (currentFileItem)
		goImage(0);
	else
		theViewer->goSplashImage();			
	
	theViewer->showMessage(mess);
}

void
selector::removeItem(selectorItem *item)
{
	cache->removeItem(item); // XXX check to see that there is no race here wrt loader?
	if (item->thumbItem)
		thumbNails->removeItem(item->thumbItem);

	if (item==currentFileItem) // fix up currentFileItem
		if (item->itemBelow()==0)
			currentFileItem=(selectorItem*)item->itemAbove();
		else
			currentFileItem=(selectorItem*)item->itemBelow();
	
	delete item;
	noItems--;
}


void
selector::configShortcuts()
{
 	rapidFileConfig cfg;
	cfg.exec();
}


int
selector::treeDirWidth()
{
 	if (opt.hideDirTree)
 		return opt.treeDirWidth;
	return treeList->width();
}

void
selector::removeMask(QPixmap *pix)
{
	QPixmap newPix;
	newPix=QPixmap(pix->size(), pix->depth());
	newPix.fill(black);
	QPainter p;
	p.begin(&newPix, this);
	p.drawPixmap(0,0, *pix);
	p.end();
	*pix=newPix;
}


void
selector::dragEnterEvent(QDragEnterEvent* event)
{
 	event->accept(QUriDrag::canDecode(event));
}

void
selector::slotDropEvent(QDropEvent *event)
{
	dropEvent(event);
}

void
selector::dropEvent(QDropEvent* event)
{
	QStringList list;
	QStrList strlist;
	if (QUriDrag::decodeLocalFiles(event, list)) {
		if (list.count()) {
			for(QStringList::Iterator it = list.begin(); it != list.end(); ++it ) {
				printf( "%s\n", (*it).latin1() );
				strlist.append((*it).latin1());
			}
		} else {
			puts("Cannot process non-local URI's yet");
			return;
		}
	} else {
		puts("invalid uri's dropped");
		return;
	}
	
	loadSelectedImages(&strlist);
	goImage(top);
}

void
selector::refileItems(QStrList &srcUris, QString destDir, QDropEvent::Action action)
{
	printf("action=%s\n",(action==QDropEvent::Move) ? "Move" : "Copy");
	
	destDir="file:"+destDir;
	if (destDir[destDir.length()-1]=='/')
		destDir.truncate(destDir.length()-1);
	//printf("Dest directory = %s\n", destDir.ascii());
	
	for(char *file=srcUris.first(); file; file=srcUris.next()) {
		puts(file);
	}

	// Stop updates while Copy/Move in progesss
	killTimer(dirWatchTimer);
	killTimer(fileWatchTimer);
	
	SyncCopyJob *s=new SyncCopyJob(KURL::List(QStringList::fromStrList(srcUris)), KURL(destDir),
																(action==QDropEvent::Move) ? KIO::CopyJob::Move :KIO::CopyJob::Copy, this);
	s->exec();
	delete s;

	dirWatchTimer=startTimer(100);
	fileWatchTimer=startTimer(5000);
	delayedUpdate=QDateTime(QDate(1900,1,1)); // Set up for immediate update
	QTimerEvent e(dirWatchTimer);
	timerEvent(&e);                           // and fake a timerEvent
}

