//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//


#ifndef SELECTOR_H
#define SELECTOR_H

#include <values.h>
#include <limits.h>
#include <pthread.h>

#include <qfileinfo.h>
#include <qdir.h>
#include <qkeycode.h>
#include <qlayout.h>
#include <qlistview.h>
#include <qmsgbox.h>
#include <qpixmap.h>
#include <qsplitter.h>
#include <qstrlist.h>
#include <qvaluelist.h>
#include <qdatetime.h>
#include <qevent.h>

#include <kaccel.h>
#include <kmenubar.h>
#include <ktmainwindow.h>
#include <kurl.h>
#include "kapp-safe.h"
#include "types.h"


class asyncLoader;
class selector;
class selectorItem;
class KDirView;
class viewer;
class thumbWidget;
class thumbWidgetItem;
class imageCache;
class QIconViewItem;

enum selectWhere {bottom=INT_MIN, randomPic=INT_MIN+1, prev=-1, next=1,
   top=INT_MAX};


class selectorItem : public QListViewItem
{
 public:
	selectorItem(QListView *parent, selectorItem *after, QFileInfo *fi);
	~selectorItem();
	virtual QString key(int column, bool ascending) const;

	QString fileName() { return fInfo.fileName(); }
	QString fileDir() { return fInfo.dirPath(); }
	QString filePath() { return fInfo.filePath(); }
	void setInfo(QFileInfo *fi);
	void updateInfo();

	QFileInfo fInfo;
	QPixmap thumbNail;
	bool loaded;
	thumbWidgetItem *thumbItem;

	static bool fullPath;
	
 private:
	void init();
};


class selector : public KTMainWindow
{
	friend class options;
	friend class optionsDialog;
	enum loadedFrom {list, directory};

	Q_OBJECT;
 public:
	selector();
	~selector();
	int  treeDirWidth();

 signals:
	void loadImage(QPixmap, QString imagePath);

public slots:
  virtual void hide();
  void fileQuit();
  void helpContents();
  void helpKeys();
  void about();
  void setStatusBarVisible(enum logicChoice set);
  void toggleStatusBar();
	void setStatus(selectorItem *item);
	void setStatusScale(float scaleFact);

	void goImage(int go);
  void newDirSelected(QListViewItem *item);
	void hideDirTree();
	void showDirTree();
	void setVisible(enum logicChoice set);
	void printImage();	
	void showImageWindow(QListViewItem *item);
	void refileImage(QString source, QString dest);
	void slotDropEvent(QDropEvent *event);
	void slotDeleteCurrentImage();

private slots:
	void slotDeleteKey();

protected:
	virtual void timerEvent(QTimerEvent *e);
  virtual void saveProperties(KConfig *conf);
  virtual void readProperties(KConfig *conf);
	virtual void resizeEvent(QResizeEvent *);
	virtual void moveEvent(QMoveEvent *);
	virtual void dragEnterEvent(QDragEnterEvent *event);
  virtual void dropEvent(QDropEvent *event);

private slots:
	void newImage(selectorItem *item);
	void newImageQLVI(QListViewItem *item);
  void options();
	void toggleDirTree();
	void setFileListView();
	void setThumbNailView();
	void thumbFileCreate();
	void thumbFileDelete();
	void thumbFileClear();
	void thumbFilePack();
	void thumbFileRebuild();
	void goFullScreen();
	void configShortcuts();
	void updateFileList(bool checkForAdditions);
	void refileItems(QStrList &srcUris, QString destDir, QDropEvent::Action action);
	void renameItem(QIconViewItem *item);
	void renameItem(QListViewItem *item);
	void deleteItems(QList<selectorItem> &items);

private:
	void loadImages(QValueList<QFileInfo> *imageList, QValueList<QDir> *dirList,
									loadedFrom where);
	void loadSelectedImages(QStrList *files);
	void loadAllImages(QString dir);
	void loadSettings();
	void getImageFormats();
	bool isImageFileName(QString file);
	void harvestImage();
	void setupLookHead(selectorItem *item);
	void removeMask(QPixmap *pix);
	void syncSelections(int dir);
	void discardLoadingImage();
	void removeItem(selectorItem *item);
	void updateItemCount();

	QString    currentDir;
	QSplitter  *split;
	QListView  *fileList;
	KDirView   *treeList;
	QBoxLayout *lay;
	QStrIList  inputFormats;
	QAccel     *accels;

  void setupMenuBar();
  void setupStatusBar();
  void readViewSetings();
	void setUpVisual();
	void cacheItem(selectorItem *item);

	viewer       *theViewer;
	thumbWidget  *thumbNails;
  QPopupMenu   *file, *view, *help, *thumb;
  KMenuBar     *menuBar;
  KStatusBar   *statusBar;
	selector     *sel;

	asyncLoader  *loader;
	pthread_t    loaderThread;
	
	imageCache   *cache;
	selectorItem *lookAheadNext;
	QString      requestedImageName;
	selectorItem *requestedItem;
	bool         loadImageFlag;

	selectorItem *currentFileItem;
	
	QValueList<QFileInfo> currentDirs;

  enum dispVisual { unknown, rgb565, rgb888x } visual;

  bool statusBarVisible, treeListFirstShow;
	int  lookAheadDepth, maxLookAhead;
	int  zoom_id, shrink_id, showStatus_id, hide_dir_id, thumb_view_id, file_view_id;
	int  noItems, preloadTimer, dirWatchTimer, fileWatchTimer;
	loadedFrom imagesLoadedFrom;
	QDateTime delayedUpdate;
};

#endif
