//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//


#include <stdlib.h>

#include <qfileinfo.h>
#include <qglobal.h>
#include <qimage.h>
#include <qdir.h>
#include <qstring.h>

#include "misc.h"
#include "thumbManager.h"
#include "kgdbm.h"
#include "options.h"
#include "asyncLoader.h"
#include "pv-jpeg.h"

thumbManager *thumbMgr;


thumbManager::thumbManager()
	: QObject()
{
	db=0;
	databaseFile="";
	databaseCurrDir="";
	requireDBWrite=false;
	
	// make a tmp filename
	char tmp[256];
	if (QString(getenv("TMPDIR")).isNull())
		strcpy(tmp, "/tmp/pikview.XXXXXX");
	else {
    strncpy(tmp, getenv("TMPDIR"), 230);
		strcat(tmp, "/pikview.XXXXXX");
	}
	if (mktemp(tmp)==0)
		tmpFileName="pikview.tmp";
	else
		tmpFileName=tmp;
}

thumbManager::~thumbManager()
{
	closeDB();
}


void
thumbManager::setUseDatabase(bool useDb)
{
	useDatabase=useDb;
}

QImage
thumbManager::getThumbNail(QString dir, QString file)
{
	QImage img;
	void *data;

	QFileInfo fi(dir+"/"+file);
	if (!fi.exists())
		return img;

	//printf("loadThumbNail: %s | %s\n",dir.data(), file.data());
	
	if (useDatabase && (databaseCurrDir!=dir)) {
		// find the correct thumbnail database to use
		//puts("checking for new database");
		QString newDatabaseFile, newDatabaseBaseDir;
		
		newDatabaseFile=findThumbNailFile(dir, newDatabaseBaseDir);
		openThumbDatabase(newDatabaseFile, newDatabaseBaseDir);
		databaseCurrDir=dir;
	}

	QString key=dir.mid(databaseBaseDir.length(), dir.length())+"/"+file;
	//printf("databaseBaseDir=%s\nkey=%s\n",databaseBaseDir.data(), key.data());
	
	bool createThumbNail=true;


	if (useDatabase && db && (db->fetchRec(key, &data)>0)) {
		int modSecs=-QFileInfo(dir+"/"+file).lastModified().secsTo(QDate(1970,1,1));
		if (modSecs==*(((int *)data)+0)) {
			int len=*(((int *)data)+1);
			if (!img.loadFromData((const uchar *)data+8, len, "JPEG")) {
				puts("Database error");
			} else
				createThumbNail=false;
		}
		free(data);
	}
	if (createThumbNail) {
		//printf("%s not in db (or no db in use) - creating\n",file.data());
		img=makeThumbNail(dir, file);
	}
	return(img);
}



QString
thumbManager::findThumbNailFile(QString startDir, QString &newDatabaseDirBase)
{
	QString newDatabaseFile="";

	QDir ddir(startDir);
	ddir.cd(ddir.canonicalPath());

	if (ddir.exists()) {
		do {
			QFileInfo thumbFile(ddir, THUMB_FILE);
			if (thumbFile.isReadable()) {
				newDatabaseFile=ddir.path()+"/"+THUMB_FILE;
				newDatabaseDirBase=ddir.path();
				break;
			}
			ddir.cdUp();
			//printf("ddir=%s\n",ddir.path().data());
		} while(ddir.path()!="/");
		if (newDatabaseFile=="") {
			newDatabaseDirBase="/"; // any dirs okay
			newDatabaseFile=QString(getenv("HOME"))+"/"+GLOBAL_THUMB_FILE;
		}
		if (newDatabaseDirBase=="/")
			newDatabaseDirBase="";
	}
	return newDatabaseFile;
}


bool
thumbManager::openThumbDatabase(QString file, QString baseDir)
{
	bool makeNewDatabase=false, ok=true;
	int version;

	// different dir but same file
	if (databaseFile==file) {
		databaseBaseDir=baseDir;
		return true;
	}
	
	//printf("openThumbDatabase %s\n",file.data());

	QFileInfo fi(file);
	if (!fi.exists() || (fi.size()==0))
		makeNewDatabase=true;

	if (file!=databaseFile) {
		if (db) delete db;
		dbReadOnly=false;
		db = new kgdbm(file, true);
		if (!db->isValid() && !requireDBWrite) {
			db = new kgdbm(file, false);
			dbReadOnly=true;
		}
		if (!db->isValid()) {
			if (!requireDBWrite)
				puts("Problem opening database file");
			closeDB();
			dbfThumbSizeX=opt.thumbSizeX;
			dbfThumbSizeY=opt.thumbSizeY;
			emit newThumbSize(dbfThumbSizeX, dbfThumbSizeY);
			return false;
		}
		databaseFile=file;
		databaseBaseDir=baseDir;
	}
	
	if (makeNewDatabase && !dbReadOnly) {
		version=1;
		db->store("#thumbSizeX", (INT32)opt.thumbSizeX, true);
		db->store("#thumbSizeY", (INT32)opt.thumbSizeY, true);
		db->store("#version", version, true);
	}
		
	dbfThumbSizeX=db->fetchInt32("#thumbSizeX");
	ok |= db->found();
	dbfThumbSizeY=db->fetchInt32("#thumbSizeX");
	ok |= db->found();
	version=db->fetchInt32("#version");
	ok |= db->found();

// 	printf("database: thumbSizeX=%d thumbSizeY=%d version=%d\n",
// 				 dbfThumbSizeX, dbfThumbSizeY, version);

	if (!ok || (version!=1)) {
		printf("*** Invalid database %s\n",file.data());
		dbfThumbSizeX=opt.thumbSizeX;
		dbfThumbSizeY=opt.thumbSizeY;
		closeDB();
		return false;
	}
	
	emit newThumbSize(dbfThumbSizeX, dbfThumbSizeY);

	return true;
}


// Makes the thumbnail (always), adds it to the database and returns
// a QImage

QImage
thumbManager::makeThumbNail(QString dir, QString file)
{
	QImage img;
	QString fileName=dir+"/"+file;

	QFileInfo fi(fileName);
	if (!fi.isReadable()) {
		return img;
	}

	QDateTime mod=fi.lastModified();
	int modSecs=-mod.secsTo(QDate(1970,1,1));

	if (QString(QImage::imageFormat(fileName)) == "JPEG") {
		img=PVJPEG::readJPEGThumbnail(fileName,
																	QSize(dbfThumbSizeX, dbfThumbSizeY));
		if (img.isNull()) {
			//printf("Invalid image %s\n",fileName.data());
			return img;
		}
	}
	else {
		QImage *nimg=imageLoad(fileName);
		img=*nimg;
		delete nimg;
		if (img.isNull()) {
			//printf("Invalid image %s\n",fileName.data());
			return img;
		}
		//printf("thumb depth=%d\n",img.depth());
		//printf("dims %d x %d\n",img.width(), img.height());
		float s=MAX((float(img.size().width())/dbfThumbSizeX),
								(float(img.size().height())/dbfThumbSizeY));
		//printf("thumb dims %d x %d\n",MAX(1,int(img.width()/s)), MAX(1,int(img.height()/s)));

		img=img.smoothScale(MAX(1,int(img.width()/s)),	
												MAX(1,int(img.height()/s)));
	}

	if (useDatabase && db && !dbReadOnly) {
		PVJPEG::jpegQuality=50;

		img.save(tmpFileName,"JPEG");
		QFile f(tmpFileName);
		int thumbSize=f.size();
		char *thumbRec=new char[thumbSize+sizeof(int)*2];

		*(((INT32*)thumbRec)+0)=(INT32)FIX_ENDIAN(modSecs);
		*(((INT32*)thumbRec)+1)=(INT32)FIX_ENDIAN(thumbSize);
		
		char *imgData=thumbRec+8;
		f.open(IO_ReadOnly);
		f.readBlock(imgData, thumbSize);
		f.close();
		f.remove();
		
		//printf("%d bytes\n",thumbSize);
		//printf("dims %d x %d\n",img.size().width(), img.size().height());
		
		QString key=dir.mid(databaseBaseDir.length(), dir.length())+"/"+file;
		//printf("save key=%s\n", key.data());
		
		db->storeData(key, thumbRec, thumbSize+sizeof(int)*2);
		delete [] thumbRec;
	}
	return(img);
}


void
thumbManager::closeDB()
{
	if (db) {
		delete db;
		db=0;
	}
	databaseFile="";
	databaseCurrDir="///";
}


void
thumbManager::packDatabase(QString database)
{
	QString dirBase, newKey;

	closeDB();

	//printf("packDatabase: %s\n", database.data());

	QFileInfo fi(database);
	if ((!fi.exists()) || (fi.size()==0))
	return;

	if (fi.fileName()==GLOBAL_THUMB_FILE)
		dirBase="";
	else
		dirBase=fi.dirPath()+"/";

	db=new kgdbm(database, true);
	
	if (!db->isValid()) {
		delete db;
		db=0;
		printf("packDatabase: cannot open database %s\n",database.data());
		return;
	}

	QString key=db->firstKey();
	while(!key.isNull()) {
		if (key.left(1)!="#") {
			fi.setFile(dirBase+key);
			//printf("Check key %s\n",key.data());
			if (!fi.exists()) {
				//printf("delete: %s\n",key.data());
				newKey=db->nextKey();
				db->deleteItem(key);
				key=newKey;
				continue;
			}
		}
		key=db->nextKey();
	}
	db->pack();
	delete db;
	db=0;
}


void
thumbManager::rebuildDatabase(QString database)
{
	QString dirBase;

	closeDB();

	QFileInfo fi(database);
	if (!fi.exists())
	return;

	if (fi.fileName()==GLOBAL_THUMB_FILE)
		dirBase="";
	else
		dirBase=fi.dirPath();

	//printf("dirBase=%s\n",dirBase.data());

	packDatabase(database);
	buildThumbNails(dirBase);
}


int
thumbManager::buildThumbNails(QString dirBase)
{
	bool useDatabaseBak=useDatabase;
	const QFileInfoList *entries;

	if (dirBase=="") // Can't/Won't rebuild the golbal thumbnail file
		return 0;

	useDatabase=true;
	requireDBWrite=true;
	
	//printf("buildThumbNails: %s\n",dirBase.data());

	QDir current(dirBase);
	if (!current.exists() || !current.isReadable()) {
		requireDBWrite=false;
		return 1;
	}
	entries=current.entryInfoList(QDir::All|QDir::Hidden, QDir::Unsorted);
	QFileInfoListIterator it(*entries);
	QFileInfo *fi;
	for(;(fi=it.current());++it) {
		if ((fi->fileName()==".") || (fi->fileName()==".."))
			continue;
		
		if (fi->isDir()) {
			//printf("found Dir: %s\n",fi->absFilePath().data());
			QFileInfo ck(fi->absFilePath()+"/"+THUMB_FILE);
			if (ck.exists()) {
				//printf("Contains its own thumb file\n");
				continue;
			}
			buildThumbNails(fi->absFilePath());
		} else {
			//printf("found file: %s\n",fi->absFilePath().data());
			getThumbNail(dirBase, fi->fileName());
		}
	}
	
	useDatabase=useDatabaseBak;
	requireDBWrite=false;
	return 0;
}
