//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//

#include <math.h>

#include <qdir.h>
#include <qpainter.h>
#include <qapp.h>
#include <qlistview.h>
#include <qdatetime.h>

#include "thumbWidget.h"
#include "thumbManager.h"
#include "options.h"
#include "misc.h"
#include "selector.h"



// =======================================================================
// Class thumbWidgetItemDrag


thumbWidgetItemDrag::thumbWidgetItemDrag(QWidget *dragSource, const char *name)
	: QIconDrag(dragSource, name)
{
}

const char*
thumbWidgetItemDrag::format(int i) const
{
	if (i==0)
		return "application/x-qiconlist";
	else if (i==1)
		return "text/uri-list";
	else
		return 0;
}

QByteArray thumbWidgetItemDrag::encodedData(const char *mime) const
{
	QByteArray a;
	if (QString(mime) == "application/x-qiconlist") {
		a = QIconDrag::encodedData(mime);
	} else if (QString(mime) == "text/uri-list") {
		QString s=urls.join("\r\n");
		a.resize(s.length());
		memcpy(a.data(), s.latin1(), s.length());
	}
	return a;
}

bool thumbWidgetItemDrag::canDecode(QMimeSource *e)
{
	return e->provides("application/x-qiconlist") || e->provides("text/uri-list");
}

void thumbWidgetItemDrag::append(const QIconDragItem &item, const QRect &pr,
																 const QRect &tr, const QString &url)
{
	QIconDrag::append(item, pr, tr);
	urls << url;
}


// =======================================================================



thumbWidgetItem::thumbWidgetItem(QIconView * parent, selectorItem *selItem )
	: QIconViewItem(parent, selItem->fileName())
{
	item=selItem;
	item->thumbItem=this;
	setRenameEnabled(true);
	setDropEnabled(false);
	if (!selItem->thumbNail.isNull())
		setPixmap(selItem->thumbNail);
}


thumbWidget::thumbWidget( QWidget *parent, const char *name, int wFlags)
	: QIconView(parent, "thumbWidget", wFlags)
{
	setAcceptDrops(false);
	
	thumbMgr=new thumbManager;
	thumbMgr->setUseDatabase(opt.useCache);
	connect(thumbMgr, SIGNAL(newThumbSize(int,int)),
					this, SLOT(updateCellSize(int,int)));

	connect(this, SIGNAL(currentChanged(QIconViewItem*)),
					this, SLOT(slotCurrentChanged(QIconViewItem*)));

	connect(this, SIGNAL(doubleClicked(QIconViewItem*)),
					this, SLOT(slotDoubleClicked(QIconViewItem*)));

	connect(this, SIGNAL(returnPressed(QIconViewItem*)),
					this, SLOT(slotDoubleClicked(QIconViewItem*)));

	maxThumbSizeX=0;
	maxThumbSizeY=0;
	selectionCount=0;
	
	setBackgroundMode( PaletteLight );
	//setTableFlags(Tbl_vScrollBar | Tbl_smoothScrolling);
	setVScrollBarMode(QScrollView::AlwaysOn);
	timerDelay=1;
	processTimer=0;

	setResizeMode(QIconView::Adjust);
	setAutoArrange(true);
	setItemsMovable(false);
	setSelectionMode(QIconView::Extended);
	setWordWrapIconText(false);
	setShowToolTips(true);
	setSorting(true);
	
	// must do this if the options dialog is run and useCache is
	// deselected
	if (!opt.useCache)
		updateCellSize(opt.thumbSizeX, opt.thumbSizeY);
}

thumbWidget::~thumbWidget()
{
	delete thumbMgr;
	opt.saveThumbnailOptions(this, true);
}

void
thumbWidget::updateCellSize(int newX, int newY)
{
	maxThumbSizeX=MAX(maxThumbSizeX, newX);
	maxThumbSizeY=MAX(maxThumbSizeY, newY);

// 	showi(maxThumbSizeX);
// 	showi(maxThumbSizeY);
	
	setGridX(maxThumbSizeX+opt.thumbPadX);
	setGridY(maxThumbSizeY+opt.thumbPadY);
	arrangeItemsInGrid(true);
}

void
thumbWidget::clear()
{
	thumbCell=0;
	if (processTimer) {
		killTimer(processTimer);
		processTimer=0;
	}	
	if (opt.useCache) {
		maxThumbSizeX=0;
		maxThumbSizeY=0;
		thumbMgr->closeDB(); // so that new thumb sizes will be emitted
	}
	QIconView::clear();
}

void
thumbWidget::setImageList(selectorItem *startItem)
{	
	clear();
	
	// Grab the list of items from the selector
	selectorItem *item;
	for(item=startItem; item; item=(selectorItem *)item->nextSibling()) {
		thumbWidgetItem *thumb=new thumbWidgetItem(this, item);
	}
	updateThumbNails();
}


void
thumbWidget::show()
{
 	if ((thumbCell!=0) && (processTimer==0))
 		processTimer=startTimer(timerDelay);
	QIconView::show();
}

void
thumbWidget::hide()
{
	if (processTimer) {
		killTimer(processTimer);
		processTimer=0;
	}
	QIconView::hide();
}

void
thumbWidget::updateThumbNails()
{
	if (processTimer) killTimer(processTimer);
	thumbCell=firstItem();
	if (isVisible() && (thumbCell)) {
		// fake a timer event to start the timer and set the cell size before
		// the first draw takes place
		processTimer=999;
		QTimerEvent e(processTimer);
		timerEvent(&e);
	}
}

void
thumbWidget::timerEvent(QTimerEvent *e)
{
	if (e->timerId() != processTimer) {
		puts("|||||||||||||| bad timer event |||||||||||||||||");
		return;
	}
	//puts("timer");
	killTimer(processTimer);
	processTimer=0;

	// Skip any thumbnails already loaded
	while(thumbCell && !((thumbWidgetItem*)thumbCell)->item->thumbNail.isNull())
		thumbCell=thumbCell->nextItem();
	if (thumbCell!=0) {
		selectorItem *item=((thumbWidgetItem*)thumbCell)->item;
		//printf("Loading: %s/%s\n",item->fileDir().latin1(), item->fileName().latin1());
		QImage img=thumbMgr->getThumbNail(item->fileDir(), item->fileName());
		if (!img.isNull()) {
			QPixmap pix;
			pix.convertFromImage(img);
			item->thumbNail=pix;
			thumbCell->setPixmap(pix);
		}
		thumbCell=thumbCell->nextItem();
	}	
	if (thumbCell==0) {
		thumbMgr->closeDB();
		arrangeItemsInGrid(true); // xxx can i remove this somehow?
	}	else
		processTimer=startTimer(timerDelay);
}


void
thumbWidget::removeItem(thumbWidgetItem *item=0)
{
	if (thumbCell==item)
		thumbCell=thumbCell->nextItem();
	delete item;
}


void
thumbWidget::slotCurrentChanged(QIconViewItem *item)
{
	// remove me? XXX
// 	printf("slotCurrentChanged %s (%d)\n",item->text().latin1(),
// 				 item->isSelected());
}

void
thumbWidget::slotDoubleClicked(QIconViewItem *item)
{
	QListViewItem *i=(QListViewItem*)((thumbWidgetItem*)item)->item;
	i->setSelected(true);
 	emit doubleClicked(i);
}

QDragObject *
thumbWidget::dragObject()
{
	if (!currentItem())
		return 0;

	QPoint orig = viewportToContents( viewport()->mapFromGlobal( QCursor::pos() ) );
	thumbWidgetItemDrag *drag = new thumbWidgetItemDrag(viewport());
// 	drag->setPixmap(*currentItem()->pixmap(),
// 									QPoint(currentItem()->pixmapRect().width()/2, currentItem()->pixmapRect().height()/2));
	for(thumbWidgetItem *item = (thumbWidgetItem*)firstItem(); item;
				item = (thumbWidgetItem*)item->nextItem() ) {
		if (item->isSelected()) {
			QIconDragItem id;
			id.setData(QCString("file:"+item->item->filePath()));
			drag->append(id,
									 QRect(item->pixmapRect(false).x() - orig.x(),
												 item->pixmapRect(false).y() - orig.y(),
												 item->pixmapRect().width(), item->pixmapRect().height()),
									 QRect(item->textRect(false).x() - orig.x(),
												 item->textRect(false).y() - orig.y(),
												 item->textRect().width(), item->textRect().height()),
									 QString("file:"+item->item->filePath()));
		}
	}

	return drag;
}
