//
// Copyright (c) 2000-2001 - Andrew Richards <ajr@users.sourceforge.net>
//
// This file is part of the PikView image viewer released under the GNU GPL.
// For license details see the file COPYING included with this distribution.
//

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <signal.h>

#include <qpainter.h>
#include <qkeycode.h>
#include <qpopupmenu.h>
#include <qbitmap.h>
#include <qdragobject.h>
#include <qtimer.h>
#include <config.h>

#include <klocale.h>

#undef Bool
#include <kconfig.h>
//#include <kwm.h>
#include "kapp-safe.h"
#include <kstddirs.h>

#include "viewer.h"
#include "selector.h"
#include "options.h"
#include "misc.h"


viewer::viewer(QStrList *files, QWidget *parent=0, const char *name=0,
							 int wFlags=0)
	: QWidget(parent, name, WResizeNoErase|WRepaintNoErase)	// use WPaintClever?
{
	setBackgroundMode(NoBackground);
	setBackgroundColor(black);
	setAcceptDrops(TRUE);

	blockPaintEvents=false;
	scaleFact=1;
	funkyTimer=0;
	infoTimer=0;
	messageTimer=0;
	slideShowTimer=0;
	slideShowRunning=false;
	imagePos=QPoint(0,0);
	winRect=QRect(0,0,0,0);
	
	if (opt.viewerSaveWinPos) {
		resize(opt.viewerWinWidth, opt.viewerWinHeight);
		move(opt.viewerWinPosX, opt.viewerWinPosY);
	}	else
		resize(opt.viewerWinWidth, opt.viewerWinHeight);
	
	if (opt.fullScreen) setFullScreen(on);
}


viewer::~viewer()
{
	if (!opt.embedded)
		opt.saveViewerOptions(this, true);
}

void
viewer::keyPressEvent(QKeyEvent *e)
{
	// Stop slideshow unless slideshow key pressed
	if (!(((e->key()==Key_S) && (e->state() | AltButton)) ||
				(e->key()==Key_Pause)))
		setSlideShow(off);

	if ((e->state()==0) && ((e->ascii()>='a' && e->ascii()<='z') ||
													(e->ascii()>='0' && e->ascii()<='9') ||
													(e->ascii()==9))) {
		if (!opt.embedded && opt.enableFiling)
			processShortcut(e->ascii());
		return;
	}

	switch(e->key()) {
	case Key_F1:      kapp->invokeHelp();
		break;
	case Key_Space:
	case Key_PageDown:emit imageSelectSignal(next);
		break;
	case Key_PageUp:  emit imageSelectSignal(prev);
		break;
	case Key_Home:  	emit imageSelectSignal(top);
		break;
	case Key_End: 	  emit imageSelectSignal(bottom);
		break;
	case Key_R:   	  if (e->state() & AltButton) emit imageSelectSignal(randomPic);
		break;
	case Key_Return:
	case Key_Enter:   if (!opt.fullScreen) emit setSelectorVisibile(toggle);
		break;
	case Key_Delete:  emit deleteCurrentImage();
		break;
	case Key_Escape:  emit viewerClosed();
		break;
	case Key_Up:
		if (e->state() & ControlButton)
			if (e->state() & ShiftButton)
				rotate180();
			else
				mirrorY();
		else
			panImage(0, -bitmapPanStep);
		break;
	case Key_Down:
		if (e->state() & ControlButton)
			if (e->state() & ShiftButton)
				rotate180();
			else
				mirrorY();
		else
			panImage(0, bitmapPanStep);
		break;
	case Key_Left:
		if (e->state() & ControlButton)
			if (e->state() & ShiftButton)
				mirrorX();
			else
				rotateLeft();
		else
			panImage(-bitmapPanStep, 0);
		break;
	case Key_Right:
		if (e->state() & ControlButton)
			if (e->state() & ShiftButton)
				mirrorX();
			else
				rotateRight();
		else
			panImage(bitmapPanStep, 0);
		break;
	case Key_Plus:		scaleImage(up);
		break;
	case Key_Minus:		scaleImage(down);
		break;
	case Key_Asterisk: setFullScreen(toggle);
		break;
	case Key_C: 			if (e->state() & AltButton) setCentering(toggle);
		break;
	case Key_Pause: 	setSlideShow(toggle);
		break;
	case Key_S: 			if (e->state() & AltButton) setSlideShow(toggle);
		break;
	case Key_H: 			if (e->state() & AltButton) setHideMouse(toggle);
		break;
	case Key_Q: 			if (e->state() & ControlButton) emit viewerClosed();
		break;
	case Key_I:
		if (e->state() & AltButton) {
			showInfo();
			repaint(infoBox, false);
		}
		break;
	case Key_1: if (e->state() & ControlButton) scaleImage(set, 1);
		break;
	case Key_2:
		if (e->state() & ControlButton) scaleImage(set, 2);
		else if (e->state() & AltButton) scaleImage(set, 0.5);
		break;
	case Key_3:
		if (e->state() & ControlButton) scaleImage(set, 3);
		else if (e->state() & AltButton) scaleImage(set, 1./3);
		break;
	case Key_4:
		if (e->state() & ControlButton) scaleImage(set, 4);
		else if (e->state() & AltButton) scaleImage(set, 0.25);
		break;
	case Key_Backspace:
		shortcut="";
// 	case Key_X:
// 		showi(pos().x());
// 		showi(pos().y());
// 		showi(opt.viewerWinPosX);
// 		showi(opt.viewerWinPosY);
// 		break;
	default:
		e->ignore();
	}
}

void
viewer::paintEvent( QPaintEvent *e )
{
	//fprintf(stderr,"paintevent\n");
	if (blockPaintEvents) {
		puts("blocked!!!");
		return;
	}
	// XXX try writing into offscreen pixmaps and then blitting them onto
  // the screen - make sure X-terminal performance doesn't suffer or
	// make blitting a user option

	// The reason this is so complicated is that you have to be *VERY* careful
	// about pixel alignment if you want to use scroll()

	if ( !currentImage.isNull() ) {         // is an image loaded?
		QPainter painter(this);
		
		// bitmapPos is 0 in a coord when the displayed image is smaller than
		// the available screen space in that dimension

		// rectangle where the scaled image lies
		QRect imageRect=QRect(QPoint(0,0), currentImage.size()*scaleFact);
		imageRect.moveBy(int(-bitmapPos.x()*scaleFact),
										 int(-bitmapPos.y()*scaleFact));
		//showRect("imageRect",imageRect);
		//showRect("geom",geometry());
		// rectangle the window sits in, the visible (displayed) part of the image
		QRect windowRect=QRect(QPoint(0,0), size());
		//showRect("windowRect",windowRect);
		// only draw the intersection of the two windows
		QRect drawRect=imageRect.intersect(windowRect);
		// and update rect
		drawRect=drawRect.intersect(e->rect());
		//showRect("drawRect",drawRect);

		// project this rectangle back on to the bitmap
		int bsx=(int)floor(drawRect.left()/scaleFact);
		int bsy=(int)floor(drawRect.top()/scaleFact);
		int bex=(int)ceil(drawRect.right()/scaleFact);
		int bey=(int)ceil(drawRect.bottom()/scaleFact);

		int sx=bitmapPos.x()+bsx;
		int sy=bitmapPos.y()+bsy;
		int sw=bex-bsx+1; // XXX why do i need the +1?
		int sh=bey-bsy+1;

		//printf("bitmap coords x=%d y=%d w=%d h=%d\n",sx,sy,sw,sh);
		//showPoint("bitmapPos",bitmapPos);

 		//printf("sf=%f\n",scaleFact);
		//showRect("updateRect",e->rect());

		
		int ox=0;
		int oy=0;
		if ((currentImage.width()*scaleFact<width())) {
			if (opt.centreWindow)
				ox=(int)((width()/scaleFact-currentImage.width())/2);
			sw=-1; // speed tweak for images < screen size which sw & sh =-1
		}
		if (currentImage.height()*scaleFact<height()) {
			if (opt.centreWindow)
				oy=(int)((height()/scaleFact-currentImage.height())/2);
			sh=-1;
		}
		
		// XXX bah - when resizing to image drawing one less pixel than required
		// fix it
		//showi(currentImage.width());
		//showi(currentImage.height());
		//showi(width());
		//showi(height());

		if (sw>=currentImage.width()) sw=-1;
		if (sh>=currentImage.height()) sh=-1;

		// only erase what is required - don't erase the image itself where
		// we are going to draw back this makes it flicker ie erase/paint
		// this feels much better
		if (!e->erased()) {
			if (!imageRect.contains(windowRect)) {
				QRect tmp;
				QRect iRect=imageRect;
				iRect.moveBy(int(ox*scaleFact), int(oy*scaleFact));
				tmp.setCoords(windowRect.left(), windowRect.top(), windowRect.right(), iRect.top());
				tmp=tmp.intersect(e->rect());
				if (tmp.isValid()) painter.eraseRect(tmp);
				tmp.setCoords(windowRect.left(), iRect.top(), iRect.left(), iRect.bottom());
				tmp=tmp.intersect(e->rect());
				if (tmp.isValid()) painter.eraseRect(tmp);
				tmp.setCoords(iRect.right(), iRect.top(), windowRect.right(), iRect.bottom());
				tmp=tmp.intersect(e->rect());
				if (tmp.isValid()) painter.eraseRect(tmp);
				tmp.setCoords(windowRect.left(), iRect.bottom(), windowRect.right(), windowRect.bottom());
				tmp=tmp.intersect(e->rect());
				if (tmp.isValid()) painter.eraseRect(tmp);
			}
		}

		if (scaleFact!=1.0) painter.scale(scaleFact,scaleFact);
		painter.drawPixmap(bsx+ox,bsy+oy, currentImage, sx,sy, sw,sh);

		//printf("drawPixmap(%d,%d, img(%d), %d,%d, %d,%d)\n",bsx+ox,bsy+oy,
		//			 (int)&currentImage, sx,sy, sw,sh);

		// this is the part of the pixmap which is rendered
		//printf("draw x=%d y=%d, sx=%d sy=%d, sw=%d sh=%d\n",
		//			 (int)(bsx*scaleFact),(int)(bsy*scaleFact), sx,sy, sw,sh);

		painter.resetXForm();

		if (infoTimer) {
			painter.fillRect(infoBox, yellow);
			painter.setPen(black);
			painter.drawRect(infoBox);
			painter.drawText(infoBox, AlignCenter, currentInfo);
		}
		if (messageTimer) {
			painter.fillRect(messageBox, yellow);
			painter.setPen(black);
			painter.drawRect(messageBox);
			painter.drawText(messageBox, AlignCenter, currentMessage);
		}
	}
}


void
viewer::timerEvent(QTimerEvent *e)
{
	if (e->timerId() == slideShowTimer)
		slideShowNext();

	if (e->timerId() == infoTimer) {
		showInfo(0);
		repaint(infoBox, false);
	}

	if (e->timerId() == messageTimer) {
		showMessage("", 0);
		repaint(messageBox, false);
	}
	if (e->timerId() == funkyTimer) {
		funkyZoom();
	}
}


void
viewer::resizeEvent( QResizeEvent * )
{
	constrainBitmapPos();
}


void
viewer::mouseMoveEvent(QMouseEvent *e)
{
	const float minPanStep=scaleFact;
	if ((e->state()==LeftButton) && (dragStartPosX+dragStartPosY!=-2.0)) {
		float diffX=e->pos().x()-dragStartPosX;
		float diffY=e->pos().y()-dragStartPosY;
		float panX=0.0, panY=0.0;
		if (fabs(diffY)>=minPanStep) {
			panY=(diffY/fabs(diffY))*floor(fabs(diffY)/minPanStep);
			dragStartPosY+=panY*minPanStep;
		}
		if (fabs(diffX)>=minPanStep) {
			panX=(diffX/fabs(diffX))*floor(fabs(diffX)/minPanStep);
			dragStartPosX+=panX*minPanStep;
		}
		panImage((int)-panX, (int)-panY);
	}
}

void
viewer::mousePressEvent( QMouseEvent *e )
{
	switch(e->button()) {
	case NoButton:
		break;

	case LeftButton:
		dragStartPosX=(float)e->pos().x();
		dragStartPosY=(float)e->pos().y();
		break;

	case MidButton:
		repaint(false);
		break;

	case RightButton:
		doContextMenu();
		break;
	}
}


void
viewer::mouseReleaseEvent( QMouseEvent *e )
{
	switch(e->button()) {
	case LeftButton:
		dragStartPosX=-1.0;
		dragStartPosY=-1.0;
		break;
	}
}


void
viewer::wheelEvent(QWheelEvent *e)
{
	e->accept();
	if (e->state() & ControlButton)
		if (e->delta()<0)
			scaleImage(up);
		else
			scaleImage(down);
	else
		if (e->delta()<0)
			emit imageSelectSignal(next);
		else
			emit imageSelectSignal(prev);	
}


void
viewer::mouseDoubleClickEvent(QMouseEvent *e)
{
	if ((currentPath=="") && (QRect(201,31, 88,44).contains(e->pos())))
		funkyZoom();
}

void
viewer::closeEvent(QCloseEvent *e)
{
	e->accept();
	emit viewerClosed();
}

void
viewer::dragEnterEvent(QDragEnterEvent* event)
{
 	event->accept(QUriDrag::canDecode(event));
}

void
viewer::dropEvent(QDropEvent* event)
{
	emit viewerDropEvent(event);
}


// focusInEvent and focusOutEvent are blank to prevent the automatic
// repainting of toplevel widgets when they receive these events
// it stops the image flickering

void
viewer::focusInEvent(QFocusEvent *)
{}

void
viewer::focusOutEvent(QFocusEvent *)
{}


void
viewer::hide()
{
	if (!isHidden() && !opt.fullScreen) {
		opt.viewerWinPosX=pos().x();
		opt.viewerWinPosY=pos().y();
	}
	QWidget::hide();
}


void
viewer::newImage(QPixmap pix, QString imagePath)
{
	currentImage=pix;
	currentPath=imagePath;
	int p=currentPath.findRev('/');
	currentType=QImage::imageFormat(imagePath);
	if (!currentImage.isNull()) {
		currentInfo=currentPath.mid(p+1, currentPath.length()) + ", " +
			QString().sprintf("%dx%d, ", currentImage.width(), currentImage.height()) + currentType;
	} else {
		currentInfo=i18n("Unable to load: ")+currentPath.mid(p+1, currentPath.length());
	}
	setCaption(currentInfo);

	// optimization off means scaling when image is smaller than the screen
	// is slower but uses less memory. Left on at the moment
	// currentImage.setOptimization(QPixmap::NoOptim);

	showInfo(0);
	if (currentImage.isNull()) {
		showInfo(100000);
		currentImage=QPixmap(300,200);
		currentImage.fill(black);
	}	else if (opt.fullScreen && opt.imageInfoInFS)
		showInfo();
	showMessage("",0);
	newImageSetup();
}

void
viewer::goSplashImage()
{
	QPixmap pix(KGlobal::dirs()->findResource("appdata", i18n("pics/splash.jpg")));

	currentImage=pix;
	currentPath="";
	currentType="";
	currentInfo=i18n("PikView")+" "+VERSION+i18n(" by Andrew Richards");
	setCaption(i18n("PikView")+" "+VERSION);
	showInfo(0);
	showMessage("",0);
	int zoomBak=opt.zoomToFit;
	opt.zoomToFit=false;
	newImageSetup();
	opt.zoomToFit=zoomBak;
	funkyTimer=startTimer((rand()%500)*1000);
}

void
viewer::newImageSetup()
{
	if (funkyTimer) {
		killTimer(funkyTimer);
		funkyTimer=0;
	}

	scaleFact=1;

	//	QRect maxWinRegion=QRect(KWM::windowRegion(KWM::currentDesktop()); XXX
	QRect maxWinRegion=QRect(QPoint(0,0), QApplication::desktop()->size());

	// winRect is the window within which the image will be displayed
	if (opt.fullScreen)
		winRect.setSize(QApplication::desktop()->size());
	else
		if (opt.resizeWindow)
			winRect.setSize(maxWinRegion.size());
		else
			winRect.setSize(size());
	//showRect("winRect",winRect);

	// Scale the image if shrinkToFit or zoomToFit is on
	//showi(opt.shrinkToFit);
	//showi(opt.zoomToFit);
	if (opt.shrinkToFit && ((currentImage.width()>winRect.width()) ||
											 (currentImage.height()>winRect.height()))) {
		float sx=(float)currentImage.width()/(float)winRect.width();
		float sy=(float)currentImage.height()/(float)winRect.height();
		scaleFact=1/MAX(sx,sy);
	} else {
		if (opt.zoomToFit && ((currentImage.width()<winRect.width()) &&
												 (currentImage.height()<winRect.height()))) {
			float sx=(float)winRect.width()/(float)currentImage.width();
			float sy=(float)winRect.height()/(float)currentImage.height();
			scaleFact=MIN(sx,sy);
		}
	}
	emit newScale(scaleFact);
	//showf(scaleFact);

	// bitmapPos is the position in bitmap coords of the displayed topLeft pixel
	bitmapPos=QPoint(0,0);
	// bitmapPanStep - pixels of the bitmap scrolled - near to 40 screen pixels
	bitmapPanStep=(int)ceil(40/scaleFact);
	//showf(scaleFact);
	//showi(bitmapPanStep);

	if (opt.resizeWindow && !opt.fullScreen) {// Resize window
		//puts("resizing win to image");
		QRect newWin=QRect(0,0, int(currentImage.width()*scaleFact), int(currentImage.height()*scaleFact));
		newWin=newWin.intersect(winRect);

		// resize will repaint but resize and repaint flickers
		if (newWin.size()!=size()) {
 			resize(newWin.size());
 			move(pos());	// why is this required (it causes flicker)
		}
	}
	repaint(false);

	if (slideShowRunning)  // restart timer
		slideShowTimer=startTimer(opt.slideShowDelay);
}



void
viewer::setFullScreen(enum logicChoice set)
{
	if (set==toggle)
		set = opt.fullScreen ? off : on;

	if (set==on) { // Set fullscreen mode
		//puts("set fullscreen");
		opt.viewerWinPosX=pos().x();
		opt.viewerWinPosY=pos().y();
		opt.viewerWinWidth=size().width();
		opt.viewerWinHeight=size().height();
		opt.fullScreen=true;
		showFullScreen();		
		newImageSetup();
	} else { // Set windowed mode
		//puts("set windowed mode");
		blockPaintEvents=true;
		hide(); // doesn't update pos/size while opt.fullScreen==true
		opt.fullScreen=false;
		showNormal();
		
		resize(opt.viewerWinWidth, opt.viewerWinHeight);
		move(opt.viewerWinPosX, opt.viewerWinPosY);
		if (opt.resizeWindow)
			newImageSetup();

		blockPaintEvents=false;
		opt.fullScreen=false;
	}
	setHideMouse(opt.hideMouse ? on : off);
}


void
viewer::setHideMouse(enum logicChoice set)
{
	if (set==toggle)
		opt.hideMouse=!opt.hideMouse;
	else
		opt.hideMouse=set;

	if (opt.fullScreen && opt.hideMouse)
		setCursor(blankCursor);
	else
		setCursor(arrowCursor);
}


void
viewer::scaleImage(enum moveDir dir, float scale=1)
{
	// Scaling craps out when the size of the virtual scaled image exceeds
	// 32767 in x or y, this is because QRect et al can't handle >32767
	// make sure this doesn't happen

	static float stepFact=exp(log(2)/4);

	QRect bitmapRect=QRect(bitmapPos, size()/scaleFact);
	float osf=scaleFact; // origional scalefactor

	if (dir==up)
		scaleFact*=stepFact;
	else if (dir==down)
		scaleFact/=stepFact;
	else if (dir==set)
		if (scaleFact==scale)
			return;
		else
			scaleFact=scale;

	if ((scaleFact<0.01) || (scaleFact>100)) {
		scaleFact=osf;
		return;
	}
	
	scaleSizeRelCenter(bitmapRect, osf/scaleFact);
	bitmapPos=bitmapRect.topLeft();

	// zoom about the centre
	constrainBitmapPos();

	// Near to 40 screen pixels
	bitmapPanStep=(int)ceil(40/scaleFact);
	//showf(scaleFact);
	//showi(bitmapPanStep);

	emit newScale(scaleFact);

	repaint(false);
}


void
viewer::constrainBitmapPos()
{
	int bpx=bitmapPos.x();
	if (((currentImage.width()-bpx)*scaleFact)<width())
		bpx=(int)(currentImage.width()-width()/scaleFact);
	if (bpx<0) bpx=0;

	bitmapPos.setX(bpx);

	int bpy=bitmapPos.y();
	if (((currentImage.height()-bpy)*scaleFact)<height())
		bpy=(int)(currentImage.height()-height()/scaleFact);
	if (bpy<0) bpy=0;
	
	bitmapPos.setY(bpy);
}


void
viewer::panImage(int px, int py)
{
	if (infoTimer) {
		showInfo(0);
		repaint(infoBox, false);
	}
	if (messageTimer) {
		showMessage("", 0);
		repaint(messageBox, false);
	}
	
	int sx=0, sy=0;
	QPoint oldBitMapPos=bitmapPos;
	bitmapPos+=QPoint(px,py);
	constrainBitmapPos();
	// the following rounding is quite deliberate
	sx=(int)(floor(oldBitMapPos.x()*scaleFact)-floor(bitmapPos.x()*scaleFact));
	sy=(int)(floor(oldBitMapPos.y()*scaleFact)-floor(bitmapPos.y()*scaleFact));
	if ((sx!=0) || (sy!=0))
		scroll(sx,sy);
}

void
viewer::setCentering(enum logicChoice set)
{
	//puts("setCentering");
	if (set==toggle)
		opt.centreWindow=!opt.centreWindow;
	else
		opt.centreWindow=set;
	repaint(false);
}


void
viewer::setSlideShow(enum logicChoice set)
{
	//puts("slide show");
	if (set==toggle)
		slideShowRunning=!slideShowRunning;
	else
		slideShowRunning=set;

	if (slideShowRunning)
		slideShowTimer=startTimer(opt.slideShowDelay);
	else
		killTimer(slideShowTimer);
}


void
viewer::slideShowNext()
{
	killTimer(slideShowTimer);

	switch(opt.slideShowDir) {
	case +1:
		emit imageSelectSignal(next);		
		break;
	case -1:
		emit imageSelectSignal(prev);
		break;
	case 0:
		emit imageSelectSignal(randomPic);
		break;
	}
}


void
viewer::showRect(char *name, QRect r)
{
	printf("Rect %s = (%d %d %d %d)\n",name, r.left(),r.top(), r.width(),r.height());
}


void
viewer::showPoint(char *name, QPoint p)
{
	printf("Point %s = (%d %d)\n",name, p.x(),p.y());
}


// Scales a rectangle by float f keeping the center point fixed
void
viewer::scaleSizeRelCenter(QRect &r, float f)
{
	//showRect("r",r);
	QPoint c=r.center();
	QSize s=r.size()*f;
	QPoint ss=QPoint(s.width(),s.height());
	
	r=QRect(c-ss/2,c+ss/2);
	//showRect("r",r);
}


QPixmap
viewer::currentPicture()
{
	return(currentImage);
}


void
viewer::doContextMenu()
{
	QPopupMenu pop, rotate;

	rotate.insertItem(i18n("Rotate left 90"), this, SLOT(rotateLeft()));
	rotate.insertItem(i18n("Rotate right 90"), this, SLOT(rotateRight()));
	rotate.insertItem(i18n("Rotate 180"), this, SLOT(rotate180()));
	rotate.insertItem(i18n("Mirror X"), this, SLOT(mirrorX()));
	rotate.insertItem(i18n("Mirror Y"), this, SLOT(mirrorY()));

	if (opt.fullScreen)
		pop.insertItem(i18n("Windowed mode"), 0);
	else
		pop.insertItem(i18n("Fullscreen"), 0);
	pop.insertItem(i18n("Transform image"), &rotate);
	int itm=pop.insertItem(i18n("Show/hide selector"), 1);
	if (opt.fullScreen)
		pop.setItemEnabled(itm, false);
	pop.insertSeparator();
	pop.insertItem(i18n("Quit"), 2);
	int res=pop.exec(QCursor::pos());

	switch (res) {
	case 0:
		setFullScreen(toggle);
		break;

	case 1:
		emit setSelectorVisibile(toggle);
		break;

	case 2:
		emit viewerClosed();
		break;
	}
}

void
viewer::rotateLeft()
{
	QPixmap pix(currentImage.height(), currentImage.width(), currentImage.depth());
	QPainter p;
	p.begin(&pix);
	p.rotate(-90);
	p.drawPixmap(-currentImage.width(),0, currentImage);
	p.end();
	newImage(pix, currentPath);
}


void
viewer::rotateRight()
{
	QPixmap pix(currentImage.height(), currentImage.width(), currentImage.depth());
	QPainter p;
	p.begin(&pix);
	p.rotate(90);
	p.drawPixmap(0, -currentImage.height(), currentImage);
	p.end();
	newImage(pix, currentPath);
}


void
viewer::rotate180()
{
	QPixmap pix(currentImage.size(), currentImage.depth());
	QPainter p;
	p.begin(&pix);
	p.rotate(180);
	p.drawPixmap(-currentImage.width(),-currentImage.height(), currentImage);
	p.end();
	newImage(pix, currentPath);
}


void
viewer::mirrorX()
{
	QPixmap pix(currentImage.size(), currentImage.depth());
	QPainter p;
	p.begin(&pix);
	p.scale(-1,1);
	p.drawPixmap(-currentImage.width(), 0, currentImage);
	p.end();
	newImage(pix, currentPath);
}


void
viewer::mirrorY()
{
	QPixmap pix(currentImage.size(), currentImage.depth());
	QPainter p;
	p.begin(&pix);
	p.scale(1,-1);
	p.drawPixmap(0, -currentImage.height(), currentImage);
	p.end();
	newImage(pix, currentPath);
}


void
viewer::showInfo(int delay=2000)
{
	if (delay==0) {
		if (infoTimer) {
			killTimer(infoTimer);
			infoTimer=0;
		}
		return;
	}
	if (infoTimer)
		killTimer(infoTimer);

	infoBox=fontMetrics().boundingRect(currentInfo);
	//printf("infoBox %s\n",currentInfo.data());
	infoBox.moveBy(5,5-infoBox.y());
	infoBox.setSize(infoBox.size()+QSize(10,10));
	
	infoTimer=startTimer(delay);
}


void
viewer::showMessage(QString mess, int delay=2000)
{
	if (delay==0) {
		if (messageTimer) {
			killTimer(messageTimer);
			messageTimer=0;
		}
		return;
	}
	if (messageTimer)
		killTimer(messageTimer);

	currentMessage=mess;
	messageBox=fontMetrics().boundingRect(currentMessage);
	messageBox.setSize(messageBox.size()+QSize(10,10));
	messageBox.moveTopLeft(QPoint(5,height()-messageBox.height()-5));
	
	messageTimer=startTimer(delay);
	repaint(messageBox, false);
}


void
viewer::processShortcut(int ch)
{
	shortcut+=(char)ch;

	if (ch==9)
		shortcut=oldShortcut;

	char *sc;
	int cnt=0;
	for(sc=opt.shortCuts.first(); sc; cnt++, sc=opt.shortCuts.next()) {
		if (strncmp(shortcut.data(), sc, shortcut.length())==0)
			break;
	}
	
	if (sc) {
		if (shortcut.length()==strlen(sc)) {
			//printf("Process: %s %s\n", sc, opt.shortCutDirs.at(cnt));
			if (currentPath!="")
				emit refileImage(currentPath, opt.shortCutDirs.at(cnt));
			oldShortcut=shortcut;
			shortcut="";
		}
	} else {
		printf("Invalid shortcut: %s\n", shortcut.data());
		shortcut="";
	}
}


void
viewer::funkyZoom()
{
	if (opt.fullScreen) return; // not made for fullscreen mode
	
	QPixmap splash(KGlobal::dirs()->findResource("appdata", i18n("pics/splash.jpg")));

	QRect target(201,31, 88,44);
	float maxZoom=MIN((float)splash.width()/target.width(), (float)splash.height()/target.height());
	//printf("maxZoom=%f\n",maxZoom);
	
	QPixmap tmp(splash.size());
	
	int 	totalSteps=15;
	float	zoom, t;
	
	for(int step=1; step<=totalSteps; step++) {
		t=(step-1.0)/(totalSteps-1);
		zoom=1.0+(maxZoom-1.0)*t*t;
		//printf("t=%f zoom=%f\n",t,zoom);
		QPainter p(&tmp, this);
		p.scale(zoom,zoom);
		p.drawPixmap(0,0, splash, int(t*target.x()), int(t*target.y()));
		p.end();
		bitBlt(this, 0, 0, &tmp);
	}
	for(int step=totalSteps+2; step>=1; step--) {
		t=(((step>totalSteps) ? totalSteps : step) -1.0)/(totalSteps-1);
		zoom=1.0+(maxZoom-1.0)*t*t;
		//printf("t=%f zoom=%f\n",t,zoom);
		QPainter p(&tmp, this);
		p.scale(zoom,zoom);
		p.drawPixmap(0,0, splash, int(t*target.x()), int(t*target.y()));
		p.end();
		bitBlt(this, 0, 0, &tmp);
	}
	if (funkyTimer)
		killTimer(funkyTimer);
	funkyTimer=startTimer((rand()%500)*1000);
}
